/****************************************************************************/
/*
                                 HprintS.c

Hyper-PRINT-Server.

CGI-compliant script.

A simple HTTP script to allow point-and-click printing of files from within 
Hypertext documents.  This is limited to specified clients for obvious reasons!
Print queues available provided to the script via C environment variables and
may be specified by assigning a symbol or defining a logical name.  Three are
required:

  1. HPRINTS_PS       list of available PostScript queues and characteristics
  2. HPRINTS_TEXT     ditto for plain text queues
  3. HPRINTS_ALLOWED  list of hosts that may print using the facility

The allowed hosts are specified as a comma-separated list of specifications.
Wildcard "*" and "%" specifications allow access to a range of hosts. Hosts may
be specified using either or both numeric addresses and host names. Hence with
the list "*.wasd.dsto.defence.gov.au,131.185.45.*" the hosts
"ws1.wasd.dsto.defence.gov.au" and "131.185.45.1", etc., can use the facility. 
Numeric addresses may be needed if DNS host name lookup is not enabled on the
server.

The list of queues is comma-separated. The "#" symbol indicates that this
print queue support the "NUMBER_UP=" ScriptPrinter parameter and is capable of
printing more than one output page per sheet. The "+" symbol indicates the
printer supports double-sided printing using the ScriptPrinter "SIDES="
parameter. Assigning symbols in a script DCL procedure:

  $ HPRINTS_TEXT = "LAS2#+,LAS3#,LAS4#,LAS5#,LAS6#,LP1,LP2,LP3"
  $ HPRINTS_PS = "PSLAS2#+,PSLAS3,PSLAS4#,PSLAS5,PSLAS6"
  $ HPRINTS_ALLOWED = "*.wasd.dsto.defence.gov.au,131.185.45.*"

or if logical names are prefered define in the process or system table:

  $ DEFINE /SYSTEM HPRINTS_TEXT "LAS2#+,LAS3#,LAS4#,LAS5#,LAS6#,LP1,LP2,LP3"
  $ DEFINE /SYSTEM HPRINTS_PS "PSLAS2#+,PSLAS3,PSLAS4#,PSLAS5,PSLAS6"
  $ DEFINE /SYSTEM HPRINTS_ALLOWED "*.wasd.dsto.defence.gov.au,131.185.45.*"


BUILD DETAILS
-------------
See BUILD_HPRINTS.COM procedure.


VERSION HISTORY
---------------
28-JUN-97  MGD  v2.2.0, updated and made slightly more generic
19-SEP-95  MGD  v2.1.1, replace <CR><LF> carriage-control with single newline
24-MAY-95  MGD  v2.1.0, minor changes for AXP compatibility
29-MAR-95  MGD  v2.0.0, modified for CGI-compliance
05-DEC-94  MGD  v1.1.0, new URL mapping functions, minor revisions
12-SEP-94  MGD  v1.0.0, initial development
*/
/****************************************************************************/

#ifdef __ALPHA
   char SoftwareID [] = "HPRINTS AXP-2.2.0";
#else
   char SoftwareID [] = "HPRINTS VAX-2.2.0";
#endif

/* standard C header files */
#include <stdio.h>
#include <ctype.h>
#include <errno.h>

/* VMS-related header files */
#include <descrip.h>
#include <libdef.h>
#include <quidef.h>
#include <rmsdef.h>
#include <sjcdef.h>
#include <ssdef.h>
#include <stsdef.h>

#define JBC$_NOSUCHQUE 294970

#ifdef __ALPHA
#   pragma nomember_alignment
#endif

#define boolean int
#define true 1
#define false 0

#define VMSok(x) (x) & STS$M_SUCCESS
#define VMSnok(x) !((x) & STS$M_SUCCESS)

/* macro provides NULL pointer if CGI variable does not exist */
#define GetCgiVarIfExists(CharPointer,CgiVariableName) \
   CharPointer = getenv(CgiVariableName)

/* macro provides pointer to empty string even if CGI variable does not exist */
#define GetCgiVar(CharPointer,CgiVariableName) \
   if ((CharPointer = getenv(CgiVariableName)) == NULL) \
       CharPointer = ""; \
   if (Debug) fprintf (stdout, "%s |%s|\n", CgiVariableName, CharPointer);

char Http200Header [] =
"HTTP/1.0 200 Document follows.\r\n\
Content-Type: text/html\r\n\
\r\n";

char Http404Header [] =
"HTTP/1.0 404 Error report follows.\r\n\
Content-Type: text/html\r\n\
\r\n";

char  Utility [] = "HPRINTS";

boolean  Debug,
         HttpHasBeenOutput;

char  *AllowedListPtr,
      *CgiFormQueuePtr,
      *CgiFormPagesPtr,
      *CgiFormSidesPtr,
      *CgiFormMenuPtr,
      *CgiPathInfoPtr,
      *CgiPathTranslatedPtr,
      *CgiRemoteAddrPtr,
      *CgiRemoteHostPtr,
      *CgiRequestMethodPtr,
      *CgiScriptNamePtr,
      *PostScriptQueueListPtr,
      *TextQueueListPtr;
      
/* required function prototypes */
char* SearchTextString (char*, char*, boolean, int*);

/****************************************************************************/
/*
*/

int main ()

{
   register int  acnt;
   register char  *cptr, *sptr;

   int  status;
   FILE  *PrintFile;

   /*********/
   /* begin */
   /*********/

   if (getenv ("HPRINTS$DBUG") != NULL) Debug = true;

   if (Debug)
      system ("show sym *");
   else
   {
      /* reopen output stream so that the '\r' and '\n' are not filtered */
#ifdef __DECC
      if ((stdout = freopen ("SYS$OUTPUT", "w", stdout, "ctx=bin")) == NULL)
         exit (vaxc$errno);
#else
      if ((stdout = freopen ("SYS$OUTPUT", "w", stdout, "rfm=udf")) == NULL)
         exit (vaxc$errno);
#endif
   }

   if ((PostScriptQueueListPtr = getenv("HPRINTS_PS")) == NULL)
   {
      ErrorGeneral ("PostScript queue list not available.", __FILE__, __LINE__);
      exit (SS$_NORMAL);
   }

   if ((TextQueueListPtr = getenv("HPRINTS_TEXT")) == NULL)
   {
      ErrorGeneral ("Plain text queue list not available.", __FILE__, __LINE__);
      exit (SS$_NORMAL);
   }

   if ((AllowedListPtr = getenv("HPRINTS_ALLOWED")) == NULL)
   {
      ErrorGeneral ("Allowed host list not available.", __FILE__, __LINE__);
      exit (SS$_NORMAL);
   }

   /*************************/
   /* get the CGI variables */
   /*************************/

   GetCgiVar (CgiRemoteAddrPtr, "WWW_REMOTE_ADDR");
   GetCgiVar (CgiRemoteHostPtr, "WWW_REMOTE_HOST");

   if (!AcceptClient ()) exit (SS$_NORMAL);

   GetCgiVar (CgiPathInfoPtr, "WWW_PATH_INFO");
   GetCgiVar (CgiPathTranslatedPtr, "WWW_PATH_TRANSLATED");
   GetCgiVar (CgiRequestMethodPtr, "WWW_REQUEST_METHOD");
   GetCgiVar (CgiScriptNamePtr, "WWW_SCRIPT_NAME");

   GetCgiVar (CgiFormQueuePtr, "WWW_FORM_QUEUE");
   GetCgiVar (CgiFormPagesPtr, "WWW_FORM_PAGES");
   GetCgiVar (CgiFormSidesPtr, "WWW_FORM_SIDES");
   GetCgiVar (CgiFormMenuPtr, "WWW_FORM_MENU");

   /***********/
   /* process */
   /***********/

   if (!CgiPathInfoPtr[0] || (CgiPathInfoPtr[0] == '/' && !CgiPathInfoPtr[1]))
   {
      ErrorGeneral ("Please specify a document path.", __FILE__, __LINE__);
      exit (SS$_NORMAL);
   }

   /* check the file exists */
   if ((PrintFile = fopen (CgiPathTranslatedPtr, "r", "shr=get")) == NULL)
   {
      status = vaxc$errno;
      ErrorVmsStatus (status, CgiPathInfoPtr, CgiPathTranslatedPtr, __FILE__, __LINE__);
      exit (SS$_NORMAL);
   }
   fclose (PrintFile);

   /*
      If a print queue parameter was supplied then it is being submitted
      for printing.  If no print queue was supplied then it is requesting
      a print queue be selected.
   */
   if (CgiFormQueuePtr[0])
      SubmitPrintJob ();
   else
   {
      if (CgiFormMenuPtr[0])
      {
         ErrorGeneral ("Please select a print queue.", __FILE__, __LINE__);
         exit (SS$_NORMAL);
      }
      SelectPrinter ();
   }

   exit (SS$_NORMAL);
}

/****************************************************************************/
/*
Check if this system belongs to any of the specified domains.  Return true if
it does.   Send an explanatory message to the client if it doesn't, and return
false.
*/

boolean AcceptClient ()

{
   register char  *cptr, *lptr, *sptr;

   char  Message [256];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AcceptClient()\n");

   /* return if anyone may use it (this could be dangerous!) */
   if (!*(lptr = AllowedListPtr)) return (true);

   while (*lptr)
   {
      sptr = lptr;
      while (*lptr && *lptr != ',') lptr++;
      if (*lptr) *lptr++ = '\0';
      /* match against host address or name */
      if (isdigit(*sptr))
         cptr = CgiRemoteAddrPtr;
      else
         cptr = CgiRemoteHostPtr;
      if (Debug) fprintf (stdout, "|%s|%s|\n", cptr, sptr);
      if (SearchTextString (cptr, sptr, false, NULL) != NULL) return (true);
   }

   sprintf (Message,
      "Sorry! &nbsp;This facility is not available to <TT>%s</TT>",
      CgiRemoteHostPtr);
   ErrorGeneral (Message, __FILE__, __LINE__);
   return (false);
}

/*****************************************************************************/
/*
String search allowing wildcard "*" (matching any multiple characters) and "%" 
(matching any single character).  Returns NULL if not found or a pointer to
start of matched string.
*/ 

char* SearchTextString
( 
register char *InThat,
register char *This,
register boolean CaseSensitive,
int *MatchedLengthPtr
)
{
/* wildcards implied at both ends of the search string */
#define IMPLIED_WILDCARDS 0

   register char  *cptr, *sptr, *inptr;
   char  *RestartCptr,
         *RestartInptr,
         *MatchPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SearchTextString()\n|%s|%s|\n", This, InThat);

   if (MatchedLengthPtr != NULL) *MatchedLengthPtr = 0;
   if (!*(cptr = This)) return (NULL);
   inptr = MatchPtr = InThat;

#if IMPLIED_WILDCARDS
   /* skip leading text up to first matching character (if any!) */
   if (*cptr != '*' && *cptr != '%')
   {
      if (CaseSensitive)
         while (*inptr && *inptr != *cptr) inptr++;
      else
         while (*inptr && toupper(*inptr) != toupper(*cptr)) inptr++;
      if (Debug && !*inptr) fprintf (stdout, "1. NOT matched!\n");
      if (!*inptr) return (NULL);
      cptr++;
      MatchPtr = inptr++;
   }
#endif /* IMPLIED_WILDCARDS */

   for (;;)
   {
      if (CaseSensitive)
      {
         while (*cptr && *inptr && *cptr == *inptr)
         {
            cptr++;
            inptr++;
         }
      }
      else
      {
         while (*cptr && *inptr && toupper(*cptr) == toupper(*inptr))
         {
            cptr++;
            inptr++;
         }
      }

#if IMPLIED_WILDCARDS
      if (!*cptr)
      {
         if (Debug) fprintf (stdout, "1. matched!\n");
         if (MatchedLengthPtr != NULL) *MatchedLengthPtr = inptr - MatchPtr;
         return (MatchPtr);
      }
#else
      if (!*cptr && !*inptr)
      {
         if (Debug) fprintf (stdout, "2. matched!\n");
         if (MatchedLengthPtr != NULL) *MatchedLengthPtr = inptr - MatchPtr;
         return (MatchPtr);
      }
      else
      if (*cptr != '*' && *cptr != '%')
         return (NULL);
#endif /* IMPLIED_WILDCARDS */

      if (Debug && !*inptr) fprintf (stdout, "3. NOT matched!\n");
      if (!*inptr) return (NULL);

      if (*cptr != '*' && *cptr != '%')
      {
         cptr = This;
         MatchPtr = ++inptr;
         continue;
      }

      if (*cptr == '%')
      {
         /* single char wildcard processing */
         if (!*inptr) break;
         cptr++;
         inptr++;
         continue;
      }

      /* asterisk wildcard matching */
      while (*cptr == '*') cptr++;

      /* an asterisk wildcard at end matches all following */
      if (!*cptr)
      {
         if (Debug) fprintf (stdout, "4. matched!\n");
         while (*inptr) inptr++;
         if (MatchedLengthPtr != NULL) *MatchedLengthPtr = inptr - MatchPtr;
         return (MatchPtr);
      }

      /* note the current position in the string (first after the wildcard) */
      RestartCptr = cptr;
      for (;;)
      {
         /* find first char in InThat matching char after wildcard */
         if (CaseSensitive)
            while (*inptr && *cptr != *inptr) inptr++;
         else
            while (*inptr && toupper(*cptr) != toupper(*inptr)) inptr++;
         /* if did not find matching char in InThat being searched */
         if (Debug && !*inptr) fprintf (stdout, "5. NOT matched!\n");
         if (!*inptr) return (NULL);
         /* note the current position in InThat being searched */
         RestartInptr = inptr;
         /* try to match the remainder of the string and InThat */
         if (CaseSensitive)
         {
            while (*cptr && *inptr && *cptr == *inptr)
            {
               cptr++;
               inptr++;
            }
         }
         else
         {
            while (*cptr && *inptr && toupper(*cptr) == toupper(*inptr))
            {
               cptr++;
               inptr++;
            }
         }
         /* if reached the end of both string and InThat - match! */
#if IMPLIED_WILDCARDS
         if (!*cptr)
         {
            if (Debug) fprintf (stdout, "6. matched!\n");
            if (MatchedLengthPtr != NULL) *MatchedLengthPtr = inptr - MatchPtr;
            return (MatchPtr);
         }
#else
         if (!*cptr && !*inptr)
         {
            if (Debug) fprintf (stdout, "7. matched!\n");
            if (MatchedLengthPtr != NULL) *MatchedLengthPtr = inptr - MatchPtr;
            return (MatchPtr);
         }
#endif /* IMPLIED_WILDCARDS */
         /* break to the external loop if we encounter another wildcard */
         if (*cptr == '*' || *cptr == '%') break;
         /* lets have another go */
         cptr = RestartCptr;
         /* starting the character following the previous attempt */
         inptr = MatchPtr = RestartInptr + 1;
      }
   }
}

/****************************************************************************/
/*
Using the <FORM> tag provide the client with a list of print queues selectable 
for the print job.
*/ 

SelectPrinter ()

{
   register char  *cptr, *sptr, *zptr;

   boolean  NumberUpSupported = false,
            SidesTwoSupported = false;
   char  Comment [256],
         PrintQueueName [256];
   char  *QueueList;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SelectPrinter()\n");

   /* determine file extension and associated queue type */
   for (cptr = sptr = CgiPathInfoPtr; *cptr; cptr++)
      if (*cptr == '/') sptr = cptr;
   while (*sptr && *sptr != '.') sptr++;
   if (strsame (sptr, ".PS", -1))
      QueueList = PostScriptQueueListPtr;
   else
      QueueList = TextQueueListPtr;

   /**************************************/
   /* HTTP header and HTML start of form */
   /**************************************/

   fprintf (stdout,
"%s\
<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2//EN\">\n\
<HTML>\n\
<HEAD>\n\
<META NAME=\"generator\" CONTENT=\"%s\">\n\
<TITLE>Printing %s</TITLE>\n\
</HEAD>\n\
<BODY>\n\
<H1>Select Print Queue</H1>\n\
<H2>Printing <TT>%s</TT></H2>\n\
<FORM ACTION=\"%s%s\">\n\
<INPUT TYPE=hidden NAME=\"menu\" VALUE=\"yes\">\n",
   Http200Header, SoftwareID,
   CgiPathInfoPtr, CgiPathInfoPtr,
   CgiScriptNamePtr, CgiPathInfoPtr);
   HttpHasBeenOutput = true;

   /****************************************************/
   /* check if any printer queues support fancy stuff! */
   /****************************************************/

   cptr = QueueList;
   while (*cptr)
   {
      sptr = PrintQueueName;
      while (*cptr && *cptr != ',' && *cptr != '#' && *cptr != '+')
          *sptr++ = *cptr++;
      *sptr = '\0';

      while (*cptr == '#' || *cptr == '+')
      {
         if (*cptr == '#') NumberUpSupported = true;
         if (*cptr == '+') SidesTwoSupported = true;
         cptr++;
      }

      /* step over the comma */
      if (*cptr) cptr++;
   }

   if (NumberUpSupported)
   {
      fprintf (stdout,
"<P>On supported printers ...\n\
<P><INPUT TYPE=checkbox NAME=\"pages\" VALUE=\"2\">\
 Check this box to <B>conserve paper</B> (2 document pages per sheet).\n");
   }

   if (NumberUpSupported && SidesTwoSupported)
      fputs ("<BR>", stdout);
   else
      fputs ("<P>On supported printers ...\n<P>", stdout);

   if (SidesTwoSupported)
   {
      fprintf (stdout,
"<INPUT TYPE=checkbox NAME=\"sides\" VALUE=\"2\">\
 Check this box to <B>print on both sides</B> of each sheet.\n");
   }

   fputs ("<P><HR SIZE=3>\n", stdout);

   /******************************************/
   /* provide a radio box selector of queues */
   /******************************************/

   cptr = QueueList;
   while (*cptr)
   {
      sptr = PrintQueueName;
      while (*cptr && *cptr != ',' && *cptr != '#' && *cptr != '+')
          *sptr++ = *cptr++;
      *sptr = '\0';

      if (*cptr == '#' || *cptr == '+')
      {
         strcpy (zptr = Comment, " <I>(supports ");
         while (*zptr) zptr++;
         while (*cptr == '#' || *cptr == '+')
         {
            if (Comment[14]) *zptr++ = ',';
            if (*cptr == '#')
            {
               strcpy (zptr, " multipage");
               while (*zptr) zptr++;
               cptr++;
               continue;
            }
            if (*cptr == '+')
            {
               strcpy (zptr, " double-sided");
               while (*zptr) zptr++;
               cptr++;
               continue;
            }
         }
         strcpy (zptr, ")</I>");
      }
      else
         Comment[0] = '\0';

      fprintf (stdout,
      "<BR><INPUT TYPE=radio NAME=\"queue\" VALUE=\"%s\"> %s%s\n",
      PrintQueueName, PrintQueueName, Comment);

      /* step over the comma */
      if (*cptr) cptr++;
   }

   /***************************/
   /* rest of HTML form, etc. */
   /***************************/

   fprintf (stdout,
"<P><HR SIZE=3>\n\
<P><INPUT TYPE=submit VALUE=\" Print \"> \
(<B>to cancel</B> this print use your browser's \
&quot;BACK&quot; button/keystroke to <B>navigate backwards</B>)\n\
</FORM>\n\
</BODY>\n\
</HTML>\n");
}

/*****************************************************************************/
/*
Using system service $SNDJBCW, submit the job for printing.
*/

SubmitPrintJob ()

{
   unsigned short  Length; 
   int  status;
   char  JobStatusOutput [256];
   struct {
      short BufferLength;
      short ItemCode;
      void  *BufferPtr;
      void  *LengthPtr;
   } items [4], NullItem = {0,0,0,0};
   int  icnt;
   struct VMSiosb {
      int  Status;
      int  DeviceInfo;
   } IOsb;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SubmitPrintJob() |%s|\n", CgiPathTranslatedPtr);

   icnt = 0;

   items[icnt].BufferLength = strlen(CgiFormQueuePtr);
   items[icnt].ItemCode = SJC$_QUEUE;
   items[icnt].BufferPtr = CgiFormQueuePtr;
   items[icnt++].LengthPtr = 0;

   if (CgiFormPagesPtr[0] == '2')
   {
      items[icnt].BufferLength = 11;
      items[icnt].ItemCode = SJC$_PARAMETER_1;
      items[icnt].BufferPtr = "NUMBER_UP=2";
      items[icnt++].LengthPtr = 0;
   }

   if (CgiFormSidesPtr[0] == '2')
   {
      items[icnt].BufferLength = 9;
      items[icnt].ItemCode = SJC$_PARAMETER_2;
      items[icnt].BufferPtr = "SIDES=TWO";
      items[icnt++].LengthPtr = 0;
   }

   items[icnt] = NullItem;

   status = sys$sndjbcw (0, SJC$_CREATE_JOB, 0, &items, &IOsb, 0, 0);

   if (VMSok (status)) status = IOsb.Status;
   if (VMSnok (status))
   {
      if (status == JBC$_NOSUCHQUE)
         ErrorVmsStatus (status, CgiFormQueuePtr, CgiPathTranslatedPtr,
                         __FILE__, __LINE__);
      else
         ErrorVmsStatus (status, CgiPathInfoPtr, CgiPathTranslatedPtr,
                         __FILE__, __LINE__);
      return (status);
   }

   icnt = 0;
   items[icnt].BufferLength = strlen(CgiPathTranslatedPtr);
   items[icnt].ItemCode = SJC$_FILE_SPECIFICATION;
   items[icnt].BufferPtr = CgiPathTranslatedPtr;
   items[icnt++].LengthPtr = 0;
   items[icnt] = NullItem;

   status = sys$sndjbcw (0, SJC$_ADD_FILE, 0, &items, 0, 0, 0);

   if (VMSok (status)) status = IOsb.Status;
   if (VMSnok (status))
   {
      ErrorVmsStatus (status, CgiPathInfoPtr, CgiPathTranslatedPtr, __FILE__, __LINE__);
      return (IOsb.Status);
   }

   icnt = 0;
   items[icnt].BufferLength = sizeof(JobStatusOutput)-1;
   items[icnt].ItemCode = SJC$_JOB_STATUS_OUTPUT;
   items[icnt].BufferPtr = JobStatusOutput;
   items[icnt++].LengthPtr = &Length;
   items[icnt] = NullItem;

   status = sys$sndjbcw (0, SJC$_CLOSE_JOB, 0, &items, 0, 0, 0);

   if (VMSok (status)) status = IOsb.Status;
   if (VMSnok (status))
   {
      ErrorVmsStatus (status, CgiPathInfoPtr, CgiPathTranslatedPtr, __FILE__, __LINE__);
      return (status);
   }

   JobStatusOutput[Length] = '\0';

   /***************************************/
   /* advise the client of the job status */
   /***************************************/

   fprintf (stdout,
"%s\
<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2//EN\">\n\
<HTML>\n\
<HEAD>\n\
<META NAME=\"generator\" CONTENT=\"%s\">\n\
<TITLE>Print Job Submitted</TITLE>\n\
</HEAD>\n\
<BODY>\n\
<H1>Print Job Submitted</H1>\n\
<H2>Printing <TT>%s</TT> to %s</H2>\n\
<PRE>%s</PRE>\n\
<P>\n\
(Do <B>not</B> use your browser's &quot;RELOAD&quot; button/keystroke, \n\
it will cause another print to be submitted!  Use your browser's\n\
&quot;BACK&quot; button/keystroke to <B>navigate backwards</B>)\n\
</BODY>\n\
</HTML>\n",
   Http200Header, SoftwareID,
   CgiPathInfoPtr, CgiPathInfoPtr,
   CgiFormQueuePtr, JobStatusOutput);
   HttpHasBeenOutput = true;

   return (status);
}

/*****************************************************************************/
/*
*/

ErrorGeneral
(
char *Text,
char *SourceFileName,
int SourceLineNumber
)
{
   register char  *cptr;

   /*********/
   /* begin */
   /*********/

   /* 
      The source file format provided by the "__FILE__" macro will be
      "device:[directory]name.type;ver".  Reduce that to "name.type".
   */
   for (cptr = SourceFileName; *cptr && *cptr != ';'; cptr++);
   if (*cptr)
   {
      while (*cptr != '.') cptr--;
      *cptr-- = '\0';
   }
   while (*cptr != ']') cptr--;
   cptr++;

   if (!HttpHasBeenOutput) fputs (Http404Header, stdout);
   fprintf (stdout,
"<!-- SoftwareID: %s Module: %s Line: %d -->\n\
<H1>ERROR!</H1>\n\
<P>Reported by server.\n\
<P>%s\n",
   SoftwareID, cptr, SourceLineNumber, Text);
}

/*****************************************************************************/
/*
*/

ErrorVmsStatus
(
int StatusValue,
char *Text,
char *HiddenText,
char *SourceFileName,
int SourceLineNumber
)
{
   static char  Message [256];
   static $DESCRIPTOR (MessageDsc, Message);

   register char  *cptr;
   int  status;
   short int  Length;

   /*********/
   /* begin */
   /*********/

   if (VMSok (status = sys$getmsg (StatusValue, &Length, &MessageDsc, 1, 0))) 
   {
      Message[Length] = '\0';
      Message[0] = toupper(Message[0]);
   }
   else
      exit (status);

   /* 
      The source file format provided by the "__FILE__" macro will be
      "device:[directory]name.type;ver".  Reduce that to "name.type".
   */
   for (cptr = SourceFileName; *cptr && *cptr != ';'; cptr++);
   if (*cptr)
   {
      while (*cptr != '.') cptr--;
      *cptr-- = '\0';
   }
   while (*cptr != ']') cptr--;
   cptr++;

   if (!HttpHasBeenOutput) fputs (Http404Header, stdout);
   fprintf (stdout,
"<!-- SoftwareID: %s Module: %s Line: %d -->\n\
<H1>ERROR!</H1>\n\
<P>Reported by server.\n\
<P>%s ... <TT>%s</TT>\n\
<!-- %%X%08.08X \"%s\" -->\n",
   SoftwareID, cptr, SourceLineNumber, Message, Text, StatusValue, HiddenText);
}

/****************************************************************************/
/*
Does a case-insensitive, character-by-character string compare and returns 
true if two strings are the same, or false if not.  If a maximum number of 
characters are specified only those will be compared, if the entire strings 
should be compared then specify the number of characters as 0.
*/ 
 
boolean strsame
(
register char *sptr1,
register char *sptr2,
register int  count
)
{
   while (*sptr1 && *sptr2)
   {
      if (toupper (*sptr1++) != toupper (*sptr2++)) return (false);
      if (count)
         if (!--count) return (true);
   }
   if (*sptr1 || *sptr2)
      return (false);
   else
      return (true);
}
 
/****************************************************************************/

