/*****************************************************************************/
/*
                                 Admin.c

Server administration support functions for HTTPd.  Many of the reports and
forms used for revision are produced by functions within the core-function
module.  The control function allow server restart, exit, abort, mapping file
reload, authorization path reload and authentication cache purging.


VERSION HISTORY
---------------
08-JUN-97  MGD  "Other" menu item, with "Clients" and "Scripting" reports,
                AdminRequestReport() function (and DclScriptingReport())
01-FEB-97  MGD  new for HTTPd version 4
*/
/*****************************************************************************/

/* standard C header files */
#include <stdio.h>
#include <ctype.h>
#include <time.h>

/* VMS related header files */
#include <descrip.h>
#include <jpidef.h>
#include <libdef.h>
#include <libdtdef.h>
#include <prvdef.h>
#include <ssdef.h>
#include <stsdef.h>

/* application related header files */
#include "wasd.h"
#include "admin.h"
#include "dcl.h"
#include "error.h"
#include "HTAdmin.h"
#include "net.h"
#include "support.h"

/********************/
/* external storage */
/********************/

extern boolean  Debug;
extern boolean  ControlExitRequested;
extern boolean  ControlRestartRequested;
extern boolean  MonitorEnabled;
extern int  CurrentConnectCount;
extern int  ExitStatus;
extern int  RequestHistoryMax;
extern int  ServerChannel;
extern int  ServerPort;
extern char  ErrorAuthorizationRequired[];
extern char  ErrorStringSize[];
extern char  HtmlSgmlDoctype[];
extern char  ServerHostName[];
extern char  SoftwareID[];
extern char  Utility[];
extern struct AccountingStruct Accounting;
extern struct AuthLoadStruct  ServerAuthLoad;
extern struct ConfigStruct  Config;
extern struct GenericListStruct  RequestList;
extern struct GenericListStruct  RequestHistoryList;

/*****************************************************************************/
/*
*/ 

int AdminBegin
(
struct RequestStruct *rqptr,
void *NextTaskFunction
)
{
   register char  *cptr, *sptr, *qptr, *zptr;

   boolean  UseServerDatabase;
   int  status;
   unsigned short  Length;
   char  Buffer [1024],
         FieldName [128],
         FieldValue [256],
         FilePath [256],
         Path [256];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AdminBegin()\n");

   if (!rqptr->AccountingDone)
      rqptr->AccountingDone = ++Accounting.DoServerAdminCount;

   if (!rqptr->RemoteUser[0])
   {
      rqptr->ResponseStatusCode = 401;
      ErrorGeneral (rqptr, ErrorAuthorizationRequired, __FILE__, __LINE__);
      RequestEnd (rqptr);
      return;
   }

   /********************************************/
   /* modules that parse their own query field */
   /********************************************/

   if (strsame (rqptr->PathInfoPtr+sizeof(HttpdInternalRevise)-1, "auth/", 5))
   {
      HTAdminBegin (rqptr, NextTaskFunction);
      return;
   }

   /**********************/
   /* parse query string */
   /**********************/

   UseServerDatabase = true;
   FilePath[0] = Path[0] = '\0';

   qptr = rqptr->QueryStringPtr;
   while (*qptr)
   {
      qptr = ParseQueryField (rqptr, qptr,
                              FieldName, sizeof(FieldName),
                              FieldValue, sizeof(FieldValue),
                              __FILE__, __LINE__);
      if (qptr == NULL)
      {
         /* error occured */
         if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
            ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
         return;
      }

      if (strsame (FieldName, "file", -1))
      {
         if (toupper(FieldValue[0]) == 'Y')
            UseServerDatabase = false;
         else
         if (toupper(FieldValue[0]) == 'N')
            UseServerDatabase = true;
      }
      else
      if (strsame (FieldName, "server", -1))
      {
         if (toupper(FieldValue[0]) == 'Y')
            UseServerDatabase = true;
         else
         if (toupper(FieldValue[0]) == 'N')
            UseServerDatabase = false;
      }
      else
      if (strsame (FieldName, "edit", -1))
         strcpy (FilePath, FieldValue);
      else
      if (strsame (FieldName, "path", -1))
         strcpy (Path, FieldValue);
      else
      {
         rqptr->ResponseStatusCode = 400;
         ErrorGeneral (rqptr, "Unknown query field.", __FILE__, __LINE__);
         if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
            ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
         return;
      }
   }

   /***************/
   /* do function */
   /***************/

   if (strsame (rqptr->PathInfoPtr, HttpdInternalAdmin, -1) ||
       strsame (rqptr->PathInfoPtr, HttpdInternalReport, -1) ||
       strsame (rqptr->PathInfoPtr, HttpdInternalRevise, -1) ||
       strsame (rqptr->PathInfoPtr, HttpdInternalControl, -1))
   {
      /********/
      /* menu */
      /********/

      AdminMenu (rqptr, NextTaskFunction);
      return;
   }
   else
   if (strsame (rqptr->PathInfoPtr, HttpdInternalReport,
                sizeof(HttpdInternalReport)-1))
   {
      /***********/
      /* reports */
      /***********/

      if (strsame (rqptr->PathInfoPtr+sizeof(HttpdInternalReport)-1,
                   "request", -1))
      {
         AdminRequestReport (rqptr, NextTaskFunction);
         return;
      }
      else
      if (strsame (rqptr->PathInfoPtr+sizeof(HttpdInternalReport)-1,
                   "config", -1) ||
          strsame (rqptr->PathInfoPtr+sizeof(HttpdInternalReport)-1,
                   "config?", 7))
      {
         ConfigReport (rqptr, NextTaskFunction, UseServerDatabase);
         return;
      }
      else
      if (strsame (rqptr->PathInfoPtr+sizeof(HttpdInternalReport)-1,
                   "paths", -1) ||
          strsame (rqptr->PathInfoPtr+sizeof(HttpdInternalReport)-1,
                   "paths?", 6))
      {
         AuthPathReport (rqptr, NextTaskFunction, UseServerDatabase);
         return;
      }
      else
      if (strsame (rqptr->PathInfoPtr+sizeof(HttpdInternalReport)-1,
                   "script", -1))
      {
         DclScriptingReport (rqptr, NextTaskFunction);
         return;
      }
      else
      if (strsame (rqptr->PathInfoPtr+sizeof(HttpdInternalReport)-1,
                   "stats", -1))
      {
         AdminReportServerStats (rqptr, NextTaskFunction);
         return;
      }
      else
      if (strsame (rqptr->PathInfoPtr+sizeof(HttpdInternalReport)-1,
                   "rules", -1) ||
          strsame (rqptr->PathInfoPtr+sizeof(HttpdInternalReport)-1,
                   "rules?", 6))
      {
         if (Path[0])
            MapUrl_ReportRuleCheck (rqptr, NextTaskFunction, Path,
                                    UseServerDatabase);
         else
            MapUrl_Report (rqptr, NextTaskFunction, UseServerDatabase);
         return;
      }
      else
      if (strsame (rqptr->PathInfoPtr+sizeof(HttpdInternalReport)-1,
                   "auth", -1) ||
          strsame (rqptr->PathInfoPtr+sizeof(HttpdInternalReport)-1,
                   "auth?", 5))
      {
         AuthCacheTreeReport (rqptr, NextTaskFunction);
         return;
      }
   }
   else
   if (strsame (rqptr->PathInfoPtr, HttpdInternalRevise,
                sizeof(HttpdInternalRevise)-1))
   {
      /**************************/
      /* configuration/revision */
      /**************************/

      if (strsame (rqptr->PathInfoPtr+sizeof(HttpdInternalRevise)-1,
                   "config", -1) ||
          strsame (rqptr->PathInfoPtr+sizeof(HttpdInternalRevise)-1,
                   "config?", 7))
      {
         if (FilePath[0])
            UpdEditFileBegin (rqptr, NextTaskFunction);
         else
            ConfigRevise (rqptr, NextTaskFunction, UseServerDatabase);
         return;
      }
      else
      if (strsame (rqptr->PathInfoPtr+sizeof(HttpdInternalRevise)-1,
                   "paths", -1) ||
          strsame (rqptr->PathInfoPtr+sizeof(HttpdInternalRevise)-1,
                   "paths?", 6))
      {
         AuthPathReport (rqptr, NextTaskFunction, UseServerDatabase);
         return;
      }
      else
      if (strsame (rqptr->PathInfoPtr+sizeof(HttpdInternalRevise)-1,
                   "rules", -1) ||
          strsame (rqptr->PathInfoPtr+sizeof(HttpdInternalRevise)-1,
                   "rules?", 6))
      {
         MapUrl_ReportRuleCheck (rqptr, NextTaskFunction, Path,
                                 UseServerDatabase);
         return;
      }
   }
   else
   if (strsame (rqptr->PathInfoPtr, HttpdInternalControl,
                Length = sizeof(HttpdInternalControl)-1))
   {
      /******************/
      /* server control */
      /******************/

      AdminControl (rqptr, NextTaskFunction, rqptr->PathInfoPtr+Length);
      return;
   }

   /********************/
   /* unknown function */
   /********************/

   rqptr->ResponseStatusCode = 403;
   ErrorGeneral (rqptr, "Unknown function.", __FILE__, __LINE__);
   if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
      ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
}

/*****************************************************************************/
/*
Provide a menu of HTTPd server administration functions.
*/ 

AdminMenu
(
struct RequestStruct *rqptr,
void *NextTaskFunction
)
{
   static $DESCRIPTOR (ResponseFaoDsc,
"!AZ\
<HTML>\n\
<HEAD>\n\
!AZ\
<TITLE>HTTPd !AZ:!UL ... Server Administration Menu</TITLE>\n\
</HEAD>\n\
<BODY>\n\
<H1>HTTPd !AZ:!UL</H1>\n\
<H2>Server Administration Menu</H2>\n\
<P>\n\
\
<TABLE CELLPADDING=5 BORDER>\n\
\
<TR><TH COLSPAN=2>Configuration</TH><TR>\n\
\
<TD COLSPAN=2 VALIGN=TOP>\n\
\
<TABLE NOBORDER>\n\
\
<TR><TD></TD><TD>|</TD>\
<TH COLSPAN=2>Report</TH><TD>|</TD>\n\
<TH COLSPAN=3>Revise</TH><TD>|</TD>\n\
<TH>Action</TH><TD>|</TD><TR>\n\
\
<TR>\n\
<TH>Statistics</TH>\n\
<TD>&nbsp;</TD>\n\
<TD>[<A HREF=\"!AZstats?\">Server</A>]</TD>\n\
<TD>&nbsp;</TD>\n\
<TD>&nbsp;</TD>\n\
<TD>&nbsp;</TD>\n\
<TD>&nbsp;</TD>\n\
<TD>&nbsp;</TD>\n\
<TD>&nbsp;</TD>\n\
<TD>[<A HREF=\"!AZzero?\">Zero</A>]</TD>\n\
</TR>\n\
\
<TR>\n\
<TH>Configuration</TH>\n\
<TD>&nbsp;</TD>\n\
<TD>[<A HREF=\"!AZconfig?server=yes\">Server</A>]</TD>\n\
<TD>[<A HREF=\"!AZconfig?file=yes\">File</A>]</TD>\n\
<TD>&nbsp;</TD>\n\
<TD>[<A HREF=\"!AZconfig?server=yes\">Server</A>]</TD>\n\
<TD>[<A HREF=\"!AZconfig?file=yes\">File</A>]</TD>\n\
<TD>[<A HREF=\"!AZ!AZconfig\">Edit</A>]</TD>\n\
</TR>\n\
\
<TR>\n\
<TH>Mapping Rules</TH>\n\
<TD>&nbsp;</TD>\n\
<TD>[<A HREF=\"!AZrules?server=yes\">Server</A>]</TD>\n\
<TD>[<A HREF=\"!AZrules?file=yes\">File</A>]</TD>\n\
<TD>&nbsp;</TD>\n\
<TD>[<Axx HREF=\"!AZrules?server=yes\">Server</Axx>]</TD>\n\
<TD>[<Axx HREF=\"!AZrules?file=yes\">File</Axx>]</TD>\n\
<TD>[<A HREF=\"!AZ!AZrules\">Edit</A>]</TD>\n\
<TD>&nbsp;</TD>\n\
<TD>[<A HREF=\"!AZrules\">Reload</A>]</TD>\n\
</TR>\n\
\
<TR>\n\
<TH>Path Authorization</TH>\n\
<TD>&nbsp;</TD>\n\
<TD>[<A HREF=\"!AZpaths?server=yes\">Server</A>]</TD>\n\
<TD>[<A HREF=\"!AZpaths?file=yes\">File</A>]</TD>\n\
<TD>&nbsp;</TD>\n\
<TD>[<Axx HREF=\"!AZpaths?server=yes\">Server</Axx>]</TD>\n\
<TD>[<Axx HREF=\"!AZpaths?file=yes\">File</Axx>]</TD>\n\
<TD>[<A HREF=\"!AZ!AZpaths\">Edit</A>]</TD>\n\
<TD>&nbsp;</TD>\n\
<TD>[<A HREF=\"!AZpaths\">Reload</A>]</TD>\n\
</TR>\n\
\
<TR>\n\
<TH>User Authentication</TH>\n\
<TD>&nbsp;</TD>\n\
<TD>[<A HREF=\"!AZauth\">Server</A>]</TD>\n\
<TD>&nbsp;</TD>\n\
<TD>&nbsp;</TD>\n\
<TD>&nbsp;</TD>\n\
<TD>[<A HREF=\"!AZauth/\">File</A>]</TD>\n\
<TD>&nbsp;</TD>\n\
<TD>&nbsp;</TD>\n\
<TD>[<A HREF=\"!AZpurge\">Purge</A>]</TD>\n\
</TR>\n\
\
<TR>\n\
<TH>Other</TH>\n\
<TD>&nbsp;</TD>\n\
<TD>[<A HREF=\"!AZrequest\">Request</A>]</TD>\n\
<TD>[<A HREF=\"!AZscript\">Script</A>]</TD>\n\
<TD>&nbsp;</TD>\n\
<TD>&nbsp;</TD>\n\
<TD>&nbsp;</TD>\n\
<TD>&nbsp;</TD>\n\
<TD>&nbsp;</TD>\n\
<TD>&nbsp;</TD>\n\
</TR>\n\
\
</TABLE>\n\
\
</TD>\n\
</TR>\n\
<TR>\n\
<TH><FONT COLOR=\"#ff0000\">Control</FONT><SUP>*</SUP></TH>\n\
\
<TD ROWSPAN=2 ALIGN=CENTER>\n\
\
<TABLE NOBORDER>\n\
<TR><TH COLSPAN=3 ALIGN=left><FONT SIZE=2>!AZ</FONT></TH></TR>\n\
\
<TR><TH ROWSPAN=2 ALIGN=left><FONT SIZE=2>Times</FONT></TH>\n\
<TH ALIGN=right><FONT SIZE=2>Up:</FONT></TH>\
<TD ALIGN=left><FONT SIZE=2>!AZ</FONT></TD></TR>\n\
\
<TR><TH ALIGN=right><FONT SIZE=2>CPU:</FONT></TH><TD ALIGN=left>\
<FONT SIZE=2>!UL !2ZL:!2ZL:!2ZL.!2ZL</FONT></TD></TR>\n\
\
<TR><TH ROWSPAN=2 ALIGN=left><FONT SIZE=2>Requests</FONT></TH>\n\
<TH ALIGN=right><FONT SIZE=2>Read:</FONT></TH>\
<TD ALIGN=left><FONT SIZE=2>!AZ</FONT></TD></TR>\n\
\
<TR><TH ALIGN=right><FONT SIZE=2>Write:</FONT></TH>\
<TD ALIGN=left><FONT SIZE=2>!AZ</FONT></TD></TR>\n\
\
<TR><TH ROWSPAN=2 ALIGN=left><FONT SIZE=2>Bytes</FONT></TH>\n\
<TH ALIGN=right><FONT SIZE=2>Rx:</FONT></TH>\
<TD ALIGN=left><FONT SIZE=2>!AZ</FONT></TD></TR>\n\
\
<TR><TH ALIGN=right><FONT SIZE=2>Tx:</FONT></TH>\
<TD ALIGN=left><FONT SIZE=2>!AZ</FONT></TD></TR>\n\
</TABLE>\n\
\
</TD>\n\
</TR>\n\
\
<TR>\n\
<TD VALIGN=TOP>\n\
\
<TABLE NOBORDER>\n\
<TR><TD><A HREF=\"!AZrestart\">Restart</A></TD></TR>\n\
<TR><TD><A HREF=\"!AZexit\">Exit</A></TD></TR>\n\
<TR><TD><A HREF=\"!AZabort\">Abort</A></TD></TR>\n\
<TR><TD><SUP>*</SUP><I>\
<FONT COLOR=\"#ff0000\">Caution</FONT>, no confirmation!!</I>\n\
</TD></TR>\n\
</TABLE>\n\
\
</TD>\n\
\
</TR>\n\
</TABLE>\n\
\
</BODY>\n\
</HTML>\n");

   static unsigned long  JpiCpuTime,
                         Pid;
   static unsigned long  ConnectTime [2],
                         CurrentTime [2],
                         JpiLoginTime [2];

   static char  UpTime [32];

   static $DESCRIPTOR (UpTimeFaoDsc, "!%D\0");
   static $DESCRIPTOR (UpTimeDsc, UpTime);

   static struct
   {
      unsigned short  buf_len;
      unsigned short  item;
      unsigned char   *buf_addr;
      unsigned short  *ret_len;
   }
      JpiItems [] =
   {
      { sizeof(JpiCpuTime), JPI$_CPUTIM, &JpiCpuTime, 0 },
      { sizeof(JpiLoginTime), JPI$_LOGINTIM, &JpiLoginTime, 0 },
      {0,0,0,0}
   };

   register char  *cptr;
   register unsigned long  *vecptr;

   int  status;
   unsigned short  Length;
   unsigned long  FaoVector [64];
   char  Buffer [6144],
         RequestsRead [16],
         RequestsWrite [16],
         BytesRx [32],
         BytesTx [32];
   $DESCRIPTOR (BufferDsc, Buffer);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AdminMenu()\n");

   if (VMSnok (status = sys$getjpiw (0, &Pid, 0, &JpiItems, 0, 0, 0)))
   {
      rqptr->ErrorTextPtr = "sys$getjpi()";
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
      return;
   }

   sys$gettim (&CurrentTime);
   lib$sub_times (&CurrentTime, &JpiLoginTime, &ConnectTime);
   sys$fao (&UpTimeFaoDsc, 0, &UpTimeDsc, &ConnectTime);

   CommaNumber (32,
                Accounting.MethodGetCount +
                Accounting.MethodHeadCount,
                RequestsRead);

   CommaNumber (32,
                Accounting.MethodDeleteCount +
                Accounting.MethodPostCount +
                Accounting.MethodPutCount,
                RequestsWrite);

   CommaNumber (64, &Accounting.QuadBytesRx[0], BytesRx);
   CommaNumber (64, &Accounting.QuadBytesTx[0], BytesTx);

   rqptr->ResponsePreExpired = PRE_EXPIRE_ADMIN;
   if ((rqptr->ResponseHeaderPtr = HttpHeader200Html (rqptr)) == NULL)
   {
      if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
      return;
   }

   vecptr = FaoVector;

   *vecptr++ = HtmlSgmlDoctype;
   *vecptr++ = HtmlMetaInfo (rqptr, NULL);

   *vecptr++ = ServerHostName;
   *vecptr++ = ServerPort;
   *vecptr++ = ServerHostName;
   *vecptr++ = ServerPort;

   /* statistics */
   *vecptr++ = HttpdInternalReport;
   *vecptr++ = HttpdInternalControl;

   /* configuration */
   *vecptr++ = HttpdInternalReport;
   *vecptr++ = HttpdInternalReport;
   *vecptr++ = HttpdInternalRevise;
   *vecptr++ = HttpdInternalRevise;
   *vecptr++ = HttpdInternalScriptUpd;
   *vecptr++ = HttpdInternalRevise;

   /* mapping */
   *vecptr++ = HttpdInternalReport;
   *vecptr++ = HttpdInternalReport;
   *vecptr++ = HttpdInternalRevise;
   *vecptr++ = HttpdInternalRevise;
   *vecptr++ = HttpdInternalScriptUpd;
   *vecptr++ = HttpdInternalRevise;
   *vecptr++ = HttpdInternalControl;

   /* authorization */
   *vecptr++ = HttpdInternalReport;
   *vecptr++ = HttpdInternalReport;
   *vecptr++ = HttpdInternalRevise;
   *vecptr++ = HttpdInternalRevise;
   *vecptr++ = HttpdInternalScriptUpd;
   *vecptr++ = HttpdInternalRevise;
   *vecptr++ = HttpdInternalControl;

   /* authentication */
   *vecptr++ = HttpdInternalReport;
   *vecptr++ = HttpdInternalRevise;
   *vecptr++ = HttpdInternalControl;

   /* other */
   *vecptr++ = HttpdInternalReport;
   *vecptr++ = HttpdInternalReport;

   /* informational accounting */
   *vecptr++ = DayDateTime (rqptr->BinaryTime, 20);
   for (cptr = UpTime; isspace(*cptr); cptr++);
   *vecptr++ = (unsigned long)cptr;
   *vecptr++ = JpiCpuTime / 8640000;             /* CPU day */
   *vecptr++ = (JpiCpuTime % 8640000) / 360000;  /* CPU hour */
   *vecptr++ = (JpiCpuTime % 360000) / 6000;     /* CPU minute */
   *vecptr++ = (JpiCpuTime % 6000 ) / 100;       /* CPU second */
   *vecptr++ = JpiCpuTime % 100;                 /* CPU 10 milliseconds */
   *vecptr++ = RequestsRead;
   *vecptr++ = RequestsWrite;
   *vecptr++ = BytesRx;
   *vecptr++ = BytesTx;

   /* server control menu items */
   *vecptr++ = HttpdInternalControl;
   *vecptr++ = HttpdInternalControl;
   *vecptr++ = HttpdInternalControl;

   status = sys$faol (&ResponseFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
      return;
   }

   Buffer[Length] = '\0';
   QioNetWrite (rqptr, 0, Buffer, Length);

   if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
      ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);

   rqptr->ResponseStatusCode = 200;
}

/*****************************************************************************/
/*
Return report on the HTTPd server's activities.  Uses blocking network write.
*/ 

AdminReportServerStats
(
struct RequestStruct *rqptr,
void *NextTaskFunction
)
{
   /* ALTPRI and SETPRI synonyms? */
#  define IMAG_PRIV_0_ALLOWED \
      (~(PRV$M_ALTPRI | PRV$M_PRMMBX | PRV$M_PSWAPM | \
         PRV$M_SYSNAM | PRV$M_SYSPRV))

#  define IMAG_PRIV_1_ALLOWED (~(0x00000000))


/*
   VAX VMS does not seem to like "!@UQ" and must be made "!@UJ"!!!
   The VAX version not supporting this doesn't seem to be documented,
   but at run-time returns a %X00000014 status (%SYSTEM-F-BADPARAM).
   Implication: VAX version can only report a maximum of 2^32-1 bytes, or
   4,294,967,295 before overflow, AXP 2^64-1 (calculator gave an ERROR :^)
*/

   static $DESCRIPTOR (ResponseFaoDsc,
"!AZ\
<HTML>\n\
<HEAD>\n\
!AZ\
<TITLE>HTTPd !AZ:!UL ... Server Statistics</TITLE>\n\
</HEAD>\n\
<BODY>\n\
<H1>HTTPd !AZ:!UL</H1>\n\
<H2>Server Statistics</H2>\n\
!AZ\n\
\
<P><TABLE CELLPADDING=2 BORDER>\n\
<TR>\
<TH COLSPAN=8>Server Process</TH></TD>\
</TR>\n\
<TR>\
<TH>Version</TH><TD COLSPAN=7>!AZ</TD>\
</TR>\n\
<TR>\
<TH>Name</TH><TD>!AZ</TD>\
<TH></TH><TH>PID<TD>!8XL</TD>\
<TH></TH><TH>User<TD>!AZ</TD>\
</TR>\n\
<TR>\
<TH>AuthPriv</TH><TD>!AZ</TD>\
<TH></TH><TH>ImagPriv</TH><TD>!AZ</TD>\
<TH></TH><TH>CurPriv</TH><TD>!AZ</TD>\
</TR>\n\
<TR>\
<TH>Startup</TH><TD>!UL</TD>\
<TH></TH><TH>Exit Status</TH><TD>%X!8XL</TD>\
<TH></TH><TH>Zeroed</TH><TD>!UL</TD>\
</TR>\n\
<TR>\
<TH>Up-time</TH><TD>!AZ</TD>\
<TH></TH><TH>CPU-time</TH><TD>!UL !2ZL:!2ZL:!2ZL.!2ZL</TD>\n\
</TR>\n\
<TH>Pg.Faults</TH><TD>!UL</TD>\n\
<TH></TH><TH>Pg.Used</TH><TD>!UL%</TD>\
</TR>\n\
<TR>\
<TH>WsSize</TH><TD>!UL (!ULKB)</TD>\
<TH></TH><TH>WsPeak</TH><TD>!UL (!ULKB)</TD>\n\
<TH></TH><TH>PeakVirt</TH><TD>!UL (!ULKB)</TD>\
</TR>\n\
<TR>\
<TH>AST</TH><TD>!UL / !UL</TD>\
<TH></TH><TH>BIO</TH><TD>!UL / !UL</TD>\
<TH></TH><TH>BYT</TH><TD>!UL / !UL</TD>\
</TR>\n\
<TR>\
<TH>DIO</TH><TD>!UL / !UL</TD>\
<TH></TH><TH>ENQ</TH><TD>!UL / !UL</TD>\
<TH></TH><TH>FILE</TH><TD>!UL / !UL</TD>\
</TR>\n\
<TR>\
<TR>\
<TH>PRC</TH><TD>!UL / !UL</TD>\
<TH></TH><TH>TQ</TH><TD>!UL / !UL</TD>\n\
</TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=2 BORDER>\n\
<TR>\
<TH COLSPAN=5>Connections</TH></TD>\
</TR>\n\
<TR>\
<TH>Total</TH><TD ALIGN=RIGHT>!UL</TD>\
<TD></TD><TH>Current</TH><TD ALIGN=RIGHT>!UL</TD>\
</TR>\n\
<TR>\
<TH>Accepted</TH><TD ALIGN=RIGHT>!UL</TD>\
<TD></TD><TH>Peak</TH><TD ALIGN=RIGHT>!UL</TD>\n\
</TR>\n\
<TR>\
<TH>Rejected</TH><TD ALIGN=RIGHT>!UL</TD>\
<TD></TD><TH>Too Busy</TH><TD ALIGN=RIGHT>!UL</TD>\
</TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=2 BORDER>\n\
<TR>\
<TH COLSPAN=5>Requests</TH></TD>\
</TR>\n\
<TR>\
<TH>Total</TH><TD>!UL</TD>\
<TD></TD><TH>Error</TH><TD>!UL</TD>\
</TR>\n\
<TR>\
<TH>Parsed</TH><TD>!UL <I>(!UL redirect)</I></TD>\
<TD></TD><TH>Forbidden</TH><TD>!UL</TD>\
</TR>\n\
<TR>\
<TH>Keep-Alive</TH><TD>!UL <I>(!UL maximum)</I></TD>\n\
<TD></TD><TH>RMS Error</TH><TD>!UL</TD>\n\
</TR>\n\
<TR>\
<TH>Pragma: no-cache</TH><TD>!UL</TD>\n\
</TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=2 BORDER>\n\
<TR><TH COLSPAN=5>HTTP Methods</TH></TD></TR>\n\
<TR>\
<TH>DELETE</TH><TH>GET</TH><TH>HEAD</TH><TH>POST</TH><TH>PUT</TH>\
</TR>\n\
<TR ALIGN=RIGHT>\
<TD>!UL</TD><TD>!UL</TD><TD>!UL</TD><TD>!UL</TD><TD>!UL</TD>\
</TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=2 BORDER>\n\
<TR><TH COLSPAN=6>Authorization</TH></TR>\n\
<TR><TH COLSPAN=3>Authentication</TH><TH></TH><TH COLSPAN=2>Paths</TH>\
<TR><TH COLSPAN=2>Failure</TH><TD ALIGN=RIGHT>!UL</TD><TH></TH>\
<TH>Yes</TH><TD ALIGN=RIGHT>!UL</TD></TR>\n\
<TR><TH ROWSPAN=3>From</TH>\
<TH>Cache</TH><TD ALIGN=RIGHT>!UL</TD><TH></TH>\
<TH>No</TH><TD ALIGN=RIGHT>!UL</TD></TR>\n\
<TR><TH><FONT SIZE=2>HTA</FONT> Database</TH><TD ALIGN=RIGHT>!UL</TD></TR>\n\
<TR><TH>VMS <FONT SIZE=2>SYSUAF</FONT></TH><TD ALIGN=RIGHT>!UL</TD></TR>\n\
<TR><TH ROWSPAN=2>Scheme</TH>\
<TH>Basic</TH><TD ALIGN=RIGHT>!UL</TD></TR>\
<TR><TH>Digest</TH><TD ALIGN=RIGHT>!UL</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=2 BORDER>\n\
<TR><TH COLSPAN=3>Server Modules</TH></TD></TR>\n\
<TR><TH COLSPAN=2>Admin</TH><TD>!UL</TD></TR>\n\
<TR><TH COLSPAN=2>Directory</TH><TD>!UL</TD></TR>\n\
<TR><TH COLSPAN=2>File</TH><TD>!UL \
&nbsp;<I>(!UL not modified)</I></TD></TR>\n\
<TR><TH COLSPAN=2>IsMap</TH><TD>!UL</TD></TR>\n\
<TR><TH COLSPAN=2>Menu</TH><TD>!UL</TD></TR>\n\
<TR><TH COLSPAN=2>Put</TH><TD>!UL</TD></TR>\n\
<TR><TH COLSPAN=2>SSI</TH><TD>!UL</TD></TR>\n\
<TR><TH COLSPAN=2>Update</TH><TD>!UL</TD></TR>\n\
<TR><TH ROWSPAN=6>DCL</TH>\
<TH>CGI <I>script</I></TH><TD>!UL</TD></TR>\n\
<TR><TH>CGIplus <I>script</I></TH>\
<TD>!UL &nbsp;<I>(!UL multiple)</I></TD></TR>\n\
<TR><TH>Auto-Scripted</TH><TD>!UL</TD></TR>\n\
<TR><TH>DCL command</TH><TD>!UL</TD></TR>\n\
<TR><TH>Spawned</TH><TD>!UL</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=2 BORDER>\n\
<TR><TH COLSPAN=8>Responses</TH></TD></TR>\n\
<TR>\
<TH>1nn</TH><TH>2nn</TH><TH>3nn</TH><TH>4nn</TH><TH>5nn</TH>\
<TH></TH><TH>Redirect</TH><TH><I>Error</I></TH>\
</TR>\n\
<TR ALIGN=RIGHT>\
<TD>!UL</TD><TD>!UL</TD><TD>!UL</TD><TD>!UL</TD><TD>!UL</TD>\
<TH></TH><TD>!UL</TD><TD>!UL</TD>\
</TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=2 BORDER>\n\
<TR><TH COLSPAN=5>Bytes</TH></TD></TR>\n\
<TR><TH>Rx</TH><TD>!AZ</TD><TH></TH><TH>Tx</TH><TD>!AZ</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=2 BORDER>\n\
<TR><TH COLSPAN=4>Failures</TH></TD></TR>\n\
<TR>\
<TH>Internal<BR>Errors</TH>\
<TH>Connect<BR><I>calloc()</I></TH>\
<TH>Heap<BR><I>calloc()</I></TH>\
<TH>Heap<BR><I>realloc()</I></TH>\
</TR>\n\
<TR ALIGN=RIGHT>\
<TD>!UL</TD><TD>!UL</TD><TD>!UL</TD><TD>!UL</TD>\
</TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=2 BORDER>\n\
<TR><TH COLSPAN=2>Other</TH></TD></TR>\n\
<TR><TH>StreamLF document conversions</TH><TD>!UL</TD></TR>\n\
</TABLE>\n\
\
</BODY>\n\
</HTML>\n");

   static $DESCRIPTOR (PrivFaoDsc,
"<CENTER><FONT COLOR=\"#ff0000\">WARNING\
<BR><FONT SIZE=-1>!XL!XL</FONT></FONT></CENTER>\0");

   static char  JpiPrcNam [16],
                JpiUserName [13],
                UpTime [32];
   static unsigned long  JpiAstCnt,
                         JpiAstLm,
                         JpiBioCnt,
                         JpiBytLm,
                         JpiBytCnt,
                         JpiBioLm,
                         JpiCpuTime,
                         JpiDioCnt,
                         JpiDioLm,
                         JpiEnqCnt,
                         JpiEnqLm,
                         JpiFilCnt,
                         JpiFilLm,
                         JpiPageFlts,
                         JpiPagFilCnt,
                         JpiPgFlQuota,
                         JpiPid,
                         JpiPrcCnt,
                         JpiPrcLm,
                         JpiTqCnt,
                         JpiTqLm,
                         JpiVirtPeak,
                         JpiWsSize,
                         JpiWsPeak,
                         Pid;
   static unsigned long  ConnectTime [2],
                         CurrentTime [2],
                         JpiAuthPriv [2],
                         JpiCurPriv [2],
                         JpiImagPriv [2],
                         JpiLoginTime [2];
   static char  LastExitStatus [16] = "n/a";
   static $DESCRIPTOR (LastExitStatusDsc, LastExitStatus);
   static $DESCRIPTOR (LastExitStatusFaoDsc, "%X!XL");
   static $DESCRIPTOR (UpTimeFaoDsc, "!%D\0");
   static $DESCRIPTOR (UpTimeDsc, UpTime);
   static struct
   {
      unsigned short  buf_len;
      unsigned short  item;
      unsigned char   *buf_addr;
      unsigned short  *ret_len;
   }
      JpiItems [] =
   {
      { sizeof(JpiAstCnt), JPI$_ASTCNT, &JpiAstCnt, 0 },
      { sizeof(JpiAstLm), JPI$_ASTLM, &JpiAstLm, 0 },
      { sizeof(JpiAuthPriv), JPI$_AUTHPRIV, &JpiAuthPriv, 0 },
      { sizeof(JpiBioCnt), JPI$_BIOCNT, &JpiBioCnt, 0 },
      { sizeof(JpiBioLm), JPI$_BIOLM, &JpiBioLm, 0 },
      { sizeof(JpiBytCnt), JPI$_BYTCNT, &JpiBytCnt, 0 },
      { sizeof(JpiBytLm), JPI$_BYTLM, &JpiBytLm, 0 },
      { sizeof(JpiCpuTime), JPI$_CPUTIM, &JpiCpuTime, 0 },
      { sizeof(JpiCurPriv), JPI$_CURPRIV, &JpiCurPriv, 0 },
      { sizeof(JpiDioCnt), JPI$_DIOCNT, &JpiDioCnt, 0 },
      { sizeof(JpiDioLm), JPI$_DIOLM, &JpiDioLm, 0 },
      { sizeof(JpiEnqCnt), JPI$_ENQCNT, &JpiEnqCnt, 0 },
      { sizeof(JpiEnqLm), JPI$_ENQLM, &JpiEnqLm, 0 },
      { sizeof(JpiFilCnt), JPI$_FILCNT, &JpiFilCnt, 0 },
      { sizeof(JpiFilLm), JPI$_FILLM, &JpiFilLm, 0 },
      { sizeof(JpiImagPriv), JPI$_IMAGPRIV, &JpiImagPriv, 0 },
      { sizeof(JpiLoginTime), JPI$_LOGINTIM, &JpiLoginTime, 0 },
      { sizeof(JpiPageFlts), JPI$_PAGEFLTS, &JpiPageFlts, 0 },
      { sizeof(JpiPagFilCnt), JPI$_PAGFILCNT, &JpiPagFilCnt, 0 },
      { sizeof(JpiPgFlQuota), JPI$_PGFLQUOTA, &JpiPgFlQuota, 0 },
      { sizeof(JpiPid), JPI$_PID, &JpiPid, 0 },
      { sizeof(JpiPrcCnt), JPI$_PRCCNT, &JpiPrcCnt, 0 },
      { sizeof(JpiPrcLm), JPI$_PRCLM, &JpiPrcLm, 0 },
      { sizeof(JpiPrcNam), JPI$_PRCNAM, &JpiPrcNam, 0 },
      { sizeof(JpiTqCnt), JPI$_TQCNT, &JpiTqCnt, 0 },
      { sizeof(JpiTqLm), JPI$_TQLM, &JpiTqLm, 0 },
      { sizeof(JpiUserName), JPI$_USERNAME, &JpiUserName, 0 },
      { sizeof(JpiVirtPeak), JPI$_VIRTPEAK, &JpiVirtPeak, 0 },
      { sizeof(JpiWsSize), JPI$_WSSIZE, &JpiWsSize, 0 },
      { sizeof(JpiWsPeak), JPI$_WSPEAK, &JpiWsPeak, 0 },
      {0,0,0,0}
   };

   register unsigned int  bit;
   register unsigned long  *vecptr;
   register char  *cptr;

   int  status;
   unsigned short  Length;
   unsigned long  FaoVector [128];
   char  AuthPrivString [128],
         CurPrivString [128],
         ImagPrivString [128],
         Buffer [8192],
         BytesRx [32],
         BytesTx [32];
   $DESCRIPTOR (BufferDsc, Buffer);
   $DESCRIPTOR (AuthPrivDsc, AuthPrivString);
   $DESCRIPTOR (ImagPrivDsc, ImagPrivString);
   $DESCRIPTOR (CurPrivDsc, CurPrivString);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AdminReportServerStats()\n");

   if (VMSnok (status = sys$getjpiw (0, &Pid, 0, &JpiItems, 0, 0, 0)))
   {
      rqptr->ErrorTextPtr = "sys$getjpi()";
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
      return;
   }

   JpiUserName[12] = '\0';
   for (cptr = JpiUserName; *cptr && *cptr != ' '; cptr++);
   *cptr = '\0';
   if (Debug) fprintf (stdout, "JpiUserName |%s|\n", JpiUserName);

   JpiPrcNam[15] = '\0';
   for (cptr = JpiPrcNam; *cptr && *cptr != ' '; cptr++);
   *cptr = '\0';
   if (Debug) fprintf (stdout, "JpiPrcNam |%s|\n", JpiPrcNam);

   sys$gettim (&CurrentTime);
   lib$sub_times (&CurrentTime, &JpiLoginTime, &ConnectTime);
   sys$fao (&UpTimeFaoDsc, 0, &UpTimeDsc, &ConnectTime);

   if (Accounting.StartupCount > 1)
   {
      sys$fao (&LastExitStatusFaoDsc, &Length, &LastExitStatusDsc, 
               Accounting.LastExitStatus);
      LastExitStatus[Length] = '\0';
   }

   CommaNumber (64, &Accounting.QuadBytesRx[0], BytesRx);
   CommaNumber (64, &Accounting.QuadBytesTx[0], BytesTx);

   if ((JpiAuthPriv[0] & ~(PRV$M_NETMBX | PRV$M_TMPMBX)) || JpiAuthPriv[1])
      sys$fao (&PrivFaoDsc, 0, &AuthPrivDsc, JpiAuthPriv[1],  JpiAuthPriv[0]);
   else
      strcpy (AuthPrivString, "as expected");

   if ((JpiImagPriv[0] & IMAG_PRIV_0_ALLOWED) ||
       (JpiImagPriv[1] & IMAG_PRIV_1_ALLOWED))
      sys$fao (&PrivFaoDsc, 0, &ImagPrivDsc, JpiImagPriv[1],  JpiImagPriv[0]);
   else
      strcpy (ImagPrivString, "as expected");

   if ((JpiCurPriv[0] & ~(PRV$M_NETMBX | PRV$M_TMPMBX)) || JpiCurPriv[1])
      sys$fao (&PrivFaoDsc, 0, &CurPrivDsc, JpiCurPriv[1],  JpiCurPriv[0]);
   else
      strcpy (CurPrivString, "as expected");

   rqptr->ResponsePreExpired = PRE_EXPIRE_ADMIN;
   if ((rqptr->ResponseHeaderPtr = HttpHeader200Html (rqptr)) == NULL)
   {
      if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
      return;
   }

   vecptr = FaoVector;

   /* META info */
   *vecptr++ = HtmlSgmlDoctype;
   *vecptr++ = HtmlMetaInfo (rqptr, NULL);

   /* page title, etc. */
   *vecptr++ = ServerHostName;
   *vecptr++ = ServerPort;
   *vecptr++ = ServerHostName;
   *vecptr++ = ServerPort;
   *vecptr++ = DayDateTime (&rqptr->BinaryTime, 20);

   /* server */
   *vecptr++ = (unsigned long)SoftwareID;
   *vecptr++ = (unsigned long)JpiPrcNam;
   *vecptr++ = JpiPid;
   *vecptr++ = (unsigned long)JpiUserName;
   *vecptr++ = AuthPrivString;
   *vecptr++ = ImagPrivString;
   *vecptr++ = CurPrivString;
   *vecptr++ = Accounting.StartupCount;
   *vecptr++ = (unsigned long)LastExitStatus;
   *vecptr++ = Accounting.ZeroedCount;
   for (cptr = UpTime; isspace(*cptr); cptr++);
   *vecptr++ = (unsigned long)cptr;
   *vecptr++ = JpiCpuTime / 8640000;             /* CPU day */
   *vecptr++ = (JpiCpuTime % 8640000) / 360000;  /* CPU hour */
   *vecptr++ = (JpiCpuTime % 360000) / 6000;     /* CPU minute */
   *vecptr++ = (JpiCpuTime % 6000 ) / 100;       /* CPU second */
   *vecptr++ = JpiCpuTime % 100;                 /* CPU 10 milliseconds */
   *vecptr++ = JpiPageFlts;
   *vecptr++ = 100 - ((JpiPagFilCnt * 100) / JpiPgFlQuota);
   *vecptr++ = JpiWsSize;
   *vecptr++ = JpiWsSize / 2;
   *vecptr++ = JpiWsPeak;
   *vecptr++ = JpiWsPeak / 2;
   *vecptr++ = JpiVirtPeak;
   *vecptr++ = JpiVirtPeak / 2;
   *vecptr++ = JpiAstLm - JpiAstCnt;
   *vecptr++ = JpiAstLm;
   *vecptr++ = JpiBioLm - JpiBioCnt;
   *vecptr++ = JpiBioLm;
   *vecptr++ = JpiBytLm - JpiBytCnt;
   *vecptr++ = JpiBytLm;
   *vecptr++ = JpiDioLm - JpiDioCnt;
   *vecptr++ = JpiDioLm;
   *vecptr++ = JpiEnqLm - JpiEnqCnt;
   *vecptr++ = JpiEnqLm;
   *vecptr++ = JpiFilLm - JpiFilCnt;
   *vecptr++ = JpiFilLm;
   *vecptr++ = JpiPrcCnt;
   *vecptr++ = JpiPrcLm;
   *vecptr++ = JpiTqLm - JpiTqCnt;
   *vecptr++ = JpiTqLm;

   /* connections */
   *vecptr++ = Accounting.ConnectCount;
   *vecptr++ = Accounting.ConnectCurrent;
   *vecptr++ = Accounting.ConnectAcceptedCount;
   *vecptr++ = Accounting.ConnectPeak;
   *vecptr++ = Accounting.ConnectRejectedCount;
   *vecptr++ = Accounting.ConnectTooBusyCount;

   /* requests */
   *vecptr++ = Accounting.RequestTotalCount;
   *vecptr++ = Accounting.RequestErrorCount;
   *vecptr++ = Accounting.RequestParseCount;
   *vecptr++ = Accounting.RedirectLocalCount;
   *vecptr++ = Accounting.RequestForbiddenCount;
   *vecptr++ = Accounting.RequestKeepAliveCount;
   *vecptr++ = Accounting.RequestKeepAliveMax;
   *vecptr++ = Accounting.RequestRmsErrorCount;
   *vecptr++ = Accounting.RequestPragmaNoCacheCount;

   /* HTTP methods */
   *vecptr++ = Accounting.MethodDeleteCount;
   *vecptr++ = Accounting.MethodGetCount;
   *vecptr++ = Accounting.MethodHeadCount;
   *vecptr++ = Accounting.MethodPostCount;
   *vecptr++ = Accounting.MethodPutCount;

   /* authorization */
   /*
      Note that the total number of paths being authorized may be
      greater than the total number of authentications.  This happens
      if a request contains both script and path authorizations,
      resulting in two authorizations for the one request!
   */
   *vecptr++ = Accounting.AuthNotAuthenticatedCount;
   *vecptr++ = Accounting.AuthAuthorizedCount;
   *vecptr++ = Accounting.AuthBasicCount +
               Accounting.AuthDigestCount -
               Accounting.AuthHtDatabaseCount -
               Accounting.AuthVmsCount;
   *vecptr++ = Accounting.AuthNotAuthorizedCount;
   *vecptr++ = Accounting.AuthHtDatabaseCount;
   *vecptr++ = Accounting.AuthVmsCount;
   *vecptr++ = Accounting.AuthBasicCount;
   *vecptr++ = Accounting.AuthDigestCount;

   /* modules */
   *vecptr++ = Accounting.DoServerAdminCount;
   *vecptr++ = Accounting.DoDirectoryCount;
   *vecptr++ = Accounting.DoFileCount;
   *vecptr++ = Accounting.DoFileNotModifiedCount;
   *vecptr++ = Accounting.DoIsMapCount;
   *vecptr++ = Accounting.DoMenuCount;
   *vecptr++ = Accounting.DoPutCount;
   *vecptr++ = Accounting.DoSsiCount;
   *vecptr++ = Accounting.DoUpdateCount;
   *vecptr++ = Accounting.DoScriptCount;
   *vecptr++ = Accounting.DoCgiPlusScriptCount;
   *vecptr++ = Accounting.DclCgiPlusReusedCount;
   *vecptr++ = Accounting.DoAutoScriptCount;
   *vecptr++ = Accounting.DoDclCommandCount;
   *vecptr++ = Accounting.DclSpawnCount;

   /* responses */
   *vecptr++ = Accounting.ResponseStatusCodeCountArray[1];
   *vecptr++ = Accounting.ResponseStatusCodeCountArray[2];
   *vecptr++ = Accounting.ResponseStatusCodeCountArray[3];
   *vecptr++ = Accounting.ResponseStatusCodeCountArray[4];
   *vecptr++ = Accounting.ResponseStatusCodeCountArray[5];
   *vecptr++ = Accounting.RedirectRemoteCount;
   *vecptr++ = Accounting.ResponseStatusCodeCountArray[0];

   /* bytes rx & tx */
   *vecptr++ = BytesRx;
   *vecptr++ = BytesTx;

   /* failures */
   *vecptr++ = Accounting.InternalErrorCount;
   *vecptr++ = Accounting.ConnectCallocFailCount;
   *vecptr++ = Accounting.HeapCallocFailCount;
   *vecptr++ = Accounting.HeapReallocFailCount;

   /* other */
   *vecptr++ = Accounting.StreamLfConversionCount;

   status = sys$faol (&ResponseFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
      return;
   }

   Buffer[Length] = '\0';
   QioNetWrite (rqptr, 0, Buffer, Length);

   if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
      ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);

   rqptr->ResponseStatusCode = 200;
}

/*****************************************************************************/
/*
Return a report on current and request history, listed most to least recent.
This function blocks while executing.
*/ 

int AdminRequestReport
(
struct RequestStruct *rqptr,
void *NextTaskFunction
)
{
   static $DESCRIPTOR (ResponseFaoDsc,
"!AZ\
<HTML>\n\
<HEAD>\n\
!AZ\
<TITLE>HTTPd !AZ:!UL ... Request Report</TITLE>\n\
</HEAD>\n\
<BODY>\n\
<H1>HTTPd !AZ:!UL</H1>\n\
<H2>Request Report</H2>\n\
!AZ\n\
\
<P><TABLE CELLPADDING=2 BORDER=1>\n\
<TR><TH COLSPAN=5><FONT SIZE=+1>Current</FONT></TH></TR>\n\
<TR><TH></TH><TH>Client</TH><TH>Time</TH><TH>Rx</TH><TH>Tx</TH></TR>\n\
<TR><TH></TH><TH COLSPAN=4>Request</TH></TR>\n\
<TR><TH COLSPAN=5></TH></TR>\n");

   static $DESCRIPTOR (CurrentRequestFaoDsc,
"<TR><TD ALIGN=right>!UL</TD>\
<TD ALIGN=left>!AZ</TD><TD><FONT SIZE=-2>!AZ</FONT></TD>\
<TD ALIGN=right>!UL</TD><TD ALIGN=right>!UL</TD></TR>\n\
<TR><TD></TD><TD ALIGN=left COLSPAN=4>\
<NOBR><TT>!AZ</TT></NOBR></TD></TR>\n");

   static $DESCRIPTOR (HistoryFaoDsc,
"</TABLE>\n\
<P><TABLE CELLPADDING=2 BORDER=1>\n\
<TR><TH COLSPAN=6><FONT SIZE=+1>History</FONT>\
 &nbsp;&nbsp;<FONT SIZE=-1>!AZ</FONT></TH></TR>\n\
<TR><TH></TH><TH>Client</TH><TH>Time</TH>\
<TH>Status</HT><TH>Rx</TH><TH>Tx</TH></TR>\n\
<TR><TH></TH><TH COLSPAN=5>Request</TH></TR>\n\
<TR><TH COLSPAN=6></TH></TR>\n");

   static $DESCRIPTOR (HistoryRequestFaoDsc,
"<TR><TD ALIGN=right>!UL</TD>\
<TD ALIGN=left>!AZ</TD><TD><FONT SIZE=-2>!AZ</FONT></TD>\
<TD ALIGN=right>!UL</TD>\
<TD ALIGN=right>!UL</TD><TD ALIGN=right>!UL</TD></TR>\n\
<TR><TD></TD><TD ALIGN=left COLSPAN=5>\
<NOBR><TT>!AZ</TT></NOBR>!AZ</TD></TR>\n");

   static $DESCRIPTOR (EndPageFaoDsc,
"</TABLE>\n\
</BODY>\n\
</HTML>\n");

   static $DESCRIPTOR (KeepAliveFaoDsc, "&quot;Keep-Alive:&quot !UL\0");
   static $DESCRIPTOR (MaximumFaoDsc, "(maximum !UL)\0");

   register int  cnt;
   register unsigned long  *vecptr;
   register char  *cptr, *sptr, *zptr;
   register struct RequestStruct  *rqeptr;
   register struct RequestHistoryStruct  *rqhptr;
   register struct GenericListEntryStruct  *leptr;

   int  status,
        Count;
   unsigned short  Length;
   unsigned long  FaoVector [32];
   char  Buffer [2048],
         Client [256],
         KeepAlive [64],
         Maximum [32],
         Request [1280],
         RequestBuffer [1024];
   $DESCRIPTOR (BufferDsc, Buffer);
   $DESCRIPTOR (KeepAliveDsc, KeepAlive);
   $DESCRIPTOR (MaximumDsc, Maximum);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AdminRequestReport()\n");

   rqptr->ResponsePreExpired = PRE_EXPIRE_ADMIN;
   if ((rqptr->ResponseHeaderPtr = HttpHeader200Html (rqptr)) == NULL)
   {
      if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
      return;
   }

   vecptr = FaoVector;

   *vecptr++ = HtmlSgmlDoctype;
   *vecptr++ = HtmlMetaInfo (rqptr, NULL);

   *vecptr++ = ServerHostName;
   *vecptr++ = ServerPort;
   *vecptr++ = ServerHostName;
   *vecptr++ = ServerPort;
   *vecptr++ = DayDateTime (&rqptr->BinaryTime, 20);

   status = sys$faol (&ResponseFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
      return;
   }

   Buffer[Length] = '\0';
   QioNetWrite (rqptr, 0, Buffer, Length);

   /********************/
   /* current requests */
   /********************/

   Count = 0;

   if (Debug)
      fprintf (stdout, "Head %d\nTail %d\n",
               RequestList.HeadPtr, RequestList.TailPtr);

   /* process the request list from least to most recent */
   for (leptr = RequestList.HeadPtr; leptr != NULL; leptr = leptr->NextPtr)
   {
      rqeptr = (struct RequestStruct*)leptr->DataPtr;

      if (Debug)
         fprintf (stdout, "%d <- %d -> %d (%d)n",
                  leptr->PrevPtr, leptr, leptr->NextPtr, rqeptr);

      vecptr = FaoVector;

      *vecptr++ = ++Count;

      zptr = (sptr = Client) + sizeof(Client);
      if (rqeptr->RemoteUser[0])
      {
         cptr = rqeptr->RemoteUser;
         while (*cptr && sptr < zptr) *sptr++ = *cptr++;
         if (sptr < zptr) *sptr++ = '@';
      }
      cptr = rqeptr->ClientHostName;
      while (*cptr && sptr < zptr) *sptr++ = *cptr++;
      if (sptr >= zptr)
         ErrorExitVmsStatus (SS$_BUFFEROVF, "client", __FILE__, __LINE__);
      *sptr = '\0';
      *vecptr++ = Client;

      *vecptr++ = DayDateTime (&rqeptr->BinaryTime, 23);
      *vecptr++ = rqeptr->BytesRx;
      *vecptr++ = rqeptr->BytesTx;

      if (rqeptr->PathInfoPtr == NULL || rqeptr->QueryStringPtr == NULL)
      {
         if (rqeptr->KeepAliveCount)
         {
            sys$fao (&KeepAliveFaoDsc, 0, &KeepAliveDsc,
                     rqeptr->KeepAliveCount);
            *vecptr++ = KeepAlive;
         }
         else
            *vecptr++ = "";
      }
      else
      {
         zptr = (sptr = Request) + sizeof(Request);
         cptr = rqeptr->HttpMethodName;
         while (*cptr && sptr < zptr) *sptr++ = *cptr++;
         if (sptr < zptr) *sptr++ = ' ';
         cptr = rqeptr->ScriptName;
         while (*cptr && sptr < zptr) *sptr++ = *cptr++;
         cptr = rqeptr->PathInfoPtr;
         while (*cptr && sptr < zptr) *sptr++ = *cptr++;
         if (rqeptr->QueryStringPtr[0] && sptr < zptr) *sptr++ = '?';
         cptr = rqeptr->QueryStringPtr;
         while (*cptr && sptr < zptr) *sptr++ = *cptr++;
         if (sptr >= zptr)
            ErrorExitVmsStatus (SS$_BUFFEROVF, "request",
                                __FILE__, __LINE__);
         *sptr = '\0';

         /* provide a line-break every 80 characters */
         cnt = -1;
         zptr = (sptr = RequestBuffer) + sizeof(RequestBuffer);
         cptr = Request;
         while (*cptr && sptr < zptr)
         {
            while (*cptr && (cnt++ % 80) && sptr < zptr) *sptr++ = *cptr++;
            if (*cptr && cnt > 1 && sptr < zptr)
            {
               *sptr++ = '\n';
               if (sptr < zptr) *sptr++ = *cptr++;
            }
         }
         if (sptr >= zptr)
            ErrorExitVmsStatus (SS$_BUFFEROVF, "request",
                                __FILE__, __LINE__);
         *sptr = '\0';

         if (CopyToHtml (Buffer, sizeof(Buffer), RequestBuffer, -1) < 0)
            ErrorExitVmsStatus (SS$_BUFFEROVF, "CopyToHtml()",
                                __FILE__, __LINE__);

         /* continue with providing a line-break every 80 characters */
         zptr = (sptr = Request) + sizeof(Request);
         cptr = Buffer;
         while (*cptr && sptr < zptr)
         {
            while (*cptr && *cptr != '\n' && sptr < zptr) *sptr++ = *cptr++;
            if (*cptr && sptr+4 < zptr)
            {
               memcpy (sptr, "<BR>", 4);
               sptr += 4;
               *cptr++;
            }
         }
         if (sptr >= zptr)
            ErrorExitVmsStatus (SS$_BUFFEROVF, "request",
                                __FILE__, __LINE__);
         *sptr = '\0';

         *vecptr++ = Request;
      }

      status = sys$faol (&CurrentRequestFaoDsc, &Length, &BufferDsc,
                         &FaoVector);
      if (VMSnok (status) || status == SS$_BUFFEROVF)
      {
         rqptr->ErrorTextPtr = "sys$faol()";
         ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
         if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
            ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
         return;
      }

      Buffer[Length] = '\0';
      QioNetWrite (rqptr, 0, Buffer, Length);
   }

   /*******************/
   /* request history */
   /*******************/

   vecptr = FaoVector;
   if (RequestHistoryMax)
   {
      sys$fao (&MaximumFaoDsc, 0, &MaximumDsc, RequestHistoryMax);
      *vecptr++ = Maximum;
   }
   else
      *vecptr++ = "<I>(disabled)</I>";

   status = sys$faol (&HistoryFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
      return;
   }

   Buffer[Length] = '\0';
   QioNetWrite (rqptr, 0, Buffer, Length);

   Count = 0;

   if (Debug)
      fprintf (stdout, "Head %d\nTail %d\n",
               RequestHistoryList.HeadPtr, RequestHistoryList.TailPtr);

   /* process the request list from least to most recent */
   for (leptr = RequestHistoryList.HeadPtr;
        leptr != NULL;
        leptr = leptr->NextPtr)
   {
      rqhptr = (struct RequestHistoryStruct*)leptr->DataPtr;

      if (Debug)
         fprintf (stdout, "%d <- %d -> %d (%d)n",
                  leptr->PrevPtr, leptr, leptr->NextPtr, rqhptr);

      vecptr = FaoVector;
      *vecptr++ = ++Count;
      *vecptr++ = rqhptr->ClientAndRequest;
      *vecptr++ = DayDateTime (&rqhptr->BinaryTime, 23);
      *vecptr++ = rqhptr->ResponseStatusCode;
      *vecptr++ = rqhptr->BytesRx;
      *vecptr++ = rqhptr->BytesTx;

      /* provide a line-break every 80 characters */
      cnt = -1;
      zptr = (sptr = RequestBuffer) + sizeof(RequestBuffer);
      cptr = rqhptr->RequestPtr;
      while (*cptr && sptr < zptr)
      {
         while (*cptr && (cnt++ % 80) && sptr < zptr) *sptr++ = *cptr++;
         if (*cptr && cnt > 1 && sptr < zptr)
         {
            *sptr++ = '\n';
            if (sptr < zptr) *sptr++ = *cptr++;
         }
      }
      if (sptr >= zptr)
         ErrorExitVmsStatus (SS$_BUFFEROVF, "request",
                             __FILE__, __LINE__);
      *sptr = '\0';

      if (CopyToHtml (Buffer, sizeof(Buffer), RequestBuffer, -1) < 0)
         ErrorExitVmsStatus (SS$_BUFFEROVF, "CopyToHtml()",
                             __FILE__, __LINE__);

      /* continue with providing a line-break every 80 characters */
      zptr = (sptr = Request) + sizeof(Request);
      cptr = Buffer;
      while (*cptr && sptr < zptr)
      {
         while (*cptr && *cptr != '\n' && sptr < zptr) *sptr++ = *cptr++;
         if (*cptr && sptr+4 < zptr)
         {
            memcpy (sptr, "<BR>", 4);
            sptr += 4;
            *cptr++;
         }
      }
      if (sptr >= zptr)
         ErrorExitVmsStatus (SS$_BUFFEROVF, "request",
                             __FILE__, __LINE__);
      *sptr = '\0';

      *vecptr++ = Request;

      if (rqhptr->Truncated)
         *vecptr++ = "[truncated!]";
      else
         *vecptr++ = "";

      status = sys$faol (&HistoryRequestFaoDsc, &Length, &BufferDsc, 
                         &FaoVector);
      if (VMSnok (status) || status == SS$_BUFFEROVF)
      {
         rqptr->ErrorTextPtr = "sys$faol()";
         ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
         if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
            ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
         return;
      }

      Buffer[Length] = '\0';
      QioNetWrite (rqptr, 0, Buffer, Length);
   }

   QioNetWrite (rqptr, 0, EndPageFaoDsc.dsc$a_pointer,
                          EndPageFaoDsc.dsc$w_length);

   if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
      ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);

   rqptr->ResponseStatusCode = 200;
}

/*****************************************************************************/
/*
*/ 

int AdminControl
(
struct RequestStruct *rqptr,
void *NextTaskFunction,
char *FunctionString
)
{
   static $DESCRIPTOR (ResponseFaoDsc,
"!AZ\
<HTML>\n\
<HEAD>\n\
!AZ\
<TITLE>Success 200</TITLE>\n\
</HEAD>\n\
<BODY>\n\
<H1>SUCCESS!!</H1>\n\
<P>Reported by server.\n\
<P>!AZServer !AZ:!UL !AZ\n\
</BODY>\n\
</HTML>\n");

   static $DESCRIPTOR (ClientsConnectedFaoDsc,
" !AZ as soon as !UL other client!AZ disconnect!AZ.</FONT>\0");

   register unsigned long  *vecptr;

   int  status;
   unsigned long  FaoVector [32];
   unsigned short  Length;
   char  *ClientsPtr,
         *DisconnectsPtr;
   char  Buffer [2048],
         ClientsConnected [128];
   $DESCRIPTOR (BufferDsc, Buffer);
   $DESCRIPTOR (ClientsConnectedDsc, ClientsConnected);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AdminControl()\n");

   if (strsame (FunctionString, "subprocess_purge", -1))
   {
      DclPurgeAllSubprocesses (rqptr, NextTaskFunction, false);
      return;
   }
   else
   if (strsame (FunctionString, "subprocess_delete", -1))
   {
      DclPurgeAllSubprocesses (rqptr, NextTaskFunction, true);
      return;
   }

   rqptr->ResponsePreExpired = PRE_EXPIRE_ADMIN;
   if ((rqptr->ResponseHeaderPtr = HttpHeader200Html (rqptr)) == NULL)
   {
      if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
      return;
   }

   vecptr = FaoVector;

   *vecptr++ = HtmlSgmlDoctype;
   *vecptr++ = HtmlMetaInfo (rqptr, NULL);

   if (strsame (FunctionString, "abort", -1))
   {
      /****************/
      /* server abort */
      /****************/

      *vecptr++ = "<FONT COLOR=\"#ff0000\">";
      *vecptr++ = ServerHostName;
      *vecptr++ = ServerPort;
      *vecptr++ = " unconditional exit!</FONT>";
      sys$faol (&ResponseFaoDsc, &Length, &BufferDsc, &FaoVector);
      Buffer[Length] = '\0';
      QioNetWrite (rqptr, 0, Buffer, Length);

      fprintf (stdout, "%%%s-I-ADMIN, %s, server abort\n",
               Utility, DateTime(NULL,20));
      ExitStatus = SS$_NORMAL;
      HttpdExit ();
      sys$delprc (0, 0);
   }
   else
   if (strsame (FunctionString, "exit", -1))
   {
      /***************/
      /* server exit */
      /***************/

      *vecptr++ = "<FONT COLOR=\"#ff0000\">";
      *vecptr++ = ServerHostName;
      *vecptr++ = ServerPort;
      if (CurrentConnectCount > 1)
      {         
         if (CurrentConnectCount > 2)
         {
            ClientsPtr = "s";
            DisconnectsPtr = "";
         }
         else
         {
            ClientsPtr = "";
            DisconnectsPtr = "s";
         }
         sys$fao (&ClientsConnectedFaoDsc, 0, &ClientsConnectedDsc,
                  "exit", CurrentConnectCount-1, ClientsPtr, DisconnectsPtr);
         *vecptr++ = ClientsConnected;
      }
      else
         *vecptr++ = " exit now!</FONT>";
      sys$faol (&ResponseFaoDsc, &Length, &BufferDsc, &FaoVector);
      Buffer[Length] = '\0';
      QioNetWrite (rqptr, 0, Buffer, Length);

      if (CurrentConnectCount > 1)
      {
         /* stop the server channel from receiving incoming requests */
         NetShutdownIncomingSocket (ServerChannel);
         ControlExitRequested = true;
         if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
            ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
         rqptr->ResponseStatusCode = 200;
         return;
      }
      else
      {
         fprintf (stdout, "%%%s-I-ADMIN, %s, server exit\n",
                  Utility, DateTime(NULL,20));
         ExitStatus = SS$_NORMAL;
         HttpdExit ();
         sys$delprc (0, 0);
      }
   }
   if (strsame (FunctionString, "restart", -1))
   {
      /******************/
      /* server restart */
      /******************/

      /* relies on the supporting DCL procedure looping immediately */
      *vecptr++ = "<FONT COLOR=\"#ff0000\">";
      *vecptr++ = ServerHostName;
      *vecptr++ = ServerPort;
      if (CurrentConnectCount > 1)
      {
         if (CurrentConnectCount > 2)
         {
            ClientsPtr = "s";
            DisconnectsPtr = "";
         }
         else
         {
            ClientsPtr = "";
            DisconnectsPtr = "s";
         }
         sys$fao (&ClientsConnectedFaoDsc, 0, &ClientsConnectedDsc,
                  "restart", CurrentConnectCount-1, ClientsPtr, DisconnectsPtr);
         *vecptr++ = ClientsConnected;
      }
      else
         *vecptr++ = "restart now!</FONT>";
      sys$faol (&ResponseFaoDsc, &Length, &BufferDsc, &FaoVector);
      Buffer[Length] = '\0';
      QioNetWrite (rqptr, 0, Buffer, Length);

      if (CurrentConnectCount > 1)
      {
         /* stop the server channel from receiving incoming requests */
         NetShutdownIncomingSocket (ServerChannel);
         ControlRestartRequested = true;
         if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
            ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
         rqptr->ResponseStatusCode = 200;
         return;
      }
      else
      {
         fprintf (stdout, "%%%s-I-ADMIN, %s, server restart\n",
                  Utility, DateTime(NULL,20));
         exit (SS$_NORMAL);
      }
   }
   else
   if (strsame (FunctionString, "rules", -1))
   {
      /************************/
      /* reload mapping rules */
      /************************/

      fprintf (stdout, "%%%s-I-ADMIN, %s, mapping rules reload\n",
               Utility, DateTime(NULL,20));

      MapUrl_ControlReload (rqptr, NextTaskFunction);
      return;
   }
   else
   if (strsame (FunctionString, "paths", -1))
   {
      /*****************************/
      /* authorization path reload */
      /*****************************/

      fprintf (stdout, "%%%s-I-ADMIN, %s, authorization paths reload\n",
               Utility, DateTime(NULL,20));

      AuthControlReload (rqptr, NextTaskFunction);
      return;
   }
   else
   if (strsame (FunctionString, "purge", -1))
   {
      /******************************/
      /* purge authentication cache */
      /******************************/

      fprintf (stdout, "%%%s-I-ADMIN, %s, purge authentication cache\n",
               Utility, DateTime(NULL,20));

      /* fake it :^) */
      rqptr->QueryStringPtr = "do=purge";
      HTAdminBegin (rqptr, NextTaskFunction);
      return;
   }
   else
   if (strsame (FunctionString, "zero", -1))
   {
      /*******************/
      /* zero accounting */
      /*******************/

      fprintf (stdout, "%%%s-I-ADMIN, %s, accounting zeroed\n",
               Utility, DateTime(NULL,20));

      memset ((char*)&Accounting + sizeof(Accounting.ZeroedCount),
              0,
              sizeof(Accounting) - sizeof(Accounting.ZeroedCount));
      Accounting.ZeroedCount++;
      if (MonitorEnabled) DefineMonitorLogicals (NULL);

      *vecptr++ = "";
      *vecptr++ = ServerHostName;
      *vecptr++ = ServerPort;
      *vecptr++ = "accounting zeroed.";
      sys$faol (&ResponseFaoDsc, &Length, &BufferDsc, &FaoVector);
      Buffer[Length] = '\0';
      QioNetWrite (rqptr, 0, Buffer, Length);

      if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
      rqptr->ResponseStatusCode = 200;
      return;
   }

   /********************/
   /* unknown function */
   /********************/

   rqptr->ResponseStatusCode = 403;
   ErrorGeneral (rqptr, "Unknown control function.", __FILE__, __LINE__);
   if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
      ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
}

/****************************************************************************/

