/*****************************************************************************/
/*
                                  HTTPd.c

WASD VMS Hypertext Transfer Protocol daemon.


OBLIGATORY COPYRIGHT NOTICE
---------------------------
WASD VMS Hypertext Services, Copyright (C) 1996,1997 Mark G.Daniel.
(prior to 01-JUL-1997, "HFRD VMS Hypertext Services")

This package is free software; you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation; version 2 of the License, or any later version. 

This package is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more 
details. 

You should have received a copy of the GNU General Public License along with 
this package; if not, write to the Free Software Foundation, Inc., 675 Mass 
Ave, Cambridge, MA 02139, USA. 


QUALIFIERS  (implemented in the CLI.c module)
----------
/ACCEPT=                comma separated list of accepted hosts/domains
/CGI_PREFIX=            prefix to CGI-script variable (symbol) names
/DBUG                   turn on all "if (Debug)" statements
/DO=                    command to be passed to server
/FILBUF=                number of bytes in record buffer
/REJECT=                comma separated list of rejected hosts/domains
/[NO]LOG[=]             logging enabled/disabled, optional log file name
/[NO]MONITOR            monitor logicals enabled/disabled
/NETBUF=                number of bytes in network read buffer
/OUTBUF=                number of bytes in output buffer
/PORT=                  server IP port number
/PRIORITY=              process priority
/[NO]PROMISCUOUS        authenticates any user name for path authorization
/SUBBUF=                number of bytes in subprocess's SYS$OUTPUT buffer
/[NO]SWAP               allow/disallow process to be swapped out
/[NO]ZERO               accounting zeroed on startup (default is non-zeroed)


VERSION HISTORY
---------------
07-JUL-97  MGD  v4.2.1, minimum heap allocation chunk size,
                        prevent keep-alive timeout redefining request logical
01-JUL-97  MGD  v4.2.0, change name to WASD (Wide Area Surveillance Division)
                        persistant DCL subprocesses and CGIplus 
                        (see re-written DCL.C module),
                        scripting and client reports,
                        potential multi-thread problems in reports fixed
27-MAR-97  MGD  v4.1.0, rationalized HTTP response header generation;
                        delete on close for "temporary" files, to support
                        UPD module "preview" functionality ... WARNING, any
                        file with a name component comprising a leading hyphen,
                        sixteen digits and a trailing hyphen will be deleted!
01-FEB-97  MGD  v4.0.0, HTTPd version 4
01-OCT-96  MGD  v3.4.0, extended server reporting  
01-AUG-96  MGD  v3.3.0, realm/path-based authorization;
                        BASIC and DIGEST authentication;
                        PUT(/POST/DELETE) module;
                        StmLf module (variable to stream-LF file conversion)
12-APR-96  MGD  v3.2.0, file record/binary now determined by record format;
                        persistent connections ("Keep-Alive" within HTTP/1.0);
                        moved RMS parse structures into thread data;
                        improved local redirection detection;
                        observed Multinet disconnection/zero-byte behaviour
                        (request now aborts if network read returns zero bytes)
15-FEB-96  MGD  v3.1.1, fixed rediculous :^( bug in 302 HTTP header;
                        minor changes to request accounting and server report;
                        minor changes for user directory support;
                        minor changes to error reporting
03-JAN-96  MGD  v3.1.0, support for both DEC TCP/IP Services and TGV MultiNet
01-DEC-95  MGD  v3.0.0, single heap for each thread's dynamic memory management;
                        extensive rework of DCL subprocess functionality;
                        HTML pre-processsing module (aka Server Side Includes);
                        NCSA/CERN compliant image-mapping module;
                        NetBufferOutput() for improving network IO;
                        miscellaneous reworks/rewrites
27-SEP-95  MGD  v2.3.0, carriage-control on non-header records from <CR><LF> 
                        to single <LF> ('\n' ... newline); some browsers expect
                        only this (e.g. Netscape 1.n was spitting on X-bitmaps)
                        added Greenwich Mean Time time-stamp functionality;
                        added 'Referer:', 'If-Modified-Since:', 'User-Agent:'
07-AUG-95  MGD  v2.2.2, optionally include commented VMS file specifications
                        in HTML documents and VMS-style directory listings
16-JUN-95  MGD  v2.2.1, added file type description to "Index of" (directory)
24-MAY-95  MGD  v2.2.0, minor changes to allow compilation on AXP platform
03-APR-95  MGD  v2.1.0, add SYSUAF authentication, POST method handling
20-DEC-94  MGD  v2.0.0, multi-threaded version
20-JUN-94  MGD  v1.0.0, single-threaded version
*/
/*****************************************************************************/

#define HttpdVersion "4.2.1"

/* if not MultiNet then UCX! */
#ifndef IP_MULTINET
#  define IP_UCX 1
#  define HttpdIp "Digital-TCPIP"
#else
#  define HttpdIp "TGV-MultiNet"
#endif

#ifdef __ALPHA
#  define HttpdArch "AXP"
#else
#  define HttpdArch "VAX"
#endif

/* standard C header files */
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

/* VMS related header files */
#include <descrip.h>
#include <prvdef.h>
#include <psldef.h>
#include <ssdef.h>
#include <stsdef.h>
#include <syidef.h>

/* application-related header files */
#include "wasd.h"
#include "cli.h"
#include "copyright.h"
#include "error.h"
#include "httpd.h"
#include "logging.h"
#include "net.h"
#include "request.h"
#include "support.h"

/******************/
/* global storage */
/******************/

/*
   These are the required privileges of the executing HTTP server.
   The server ACCOUNT should only have TMPMBX and NETMBX (just for
   extra security, policy ... keep privileged accounts to a minimum).
   The image should be installed with SYSPRV, PRMMBX, PSWAPM and SYSNAM.
   Subprocesses are spawned only with the process's current privileges,
   which are always maintained at TMPMBX and NETMBX.  If additional
   privileges are required for any particular purpose (e.g. binding to
   a privileged IP port) then they are enabled, the action performed,
   and then they are disabled again immediately.
*/
unsigned long  AvJoePrivMask [2] = { PRV$M_NETMBX | PRV$M_TMPMBX, 0 },
               ResetPrivMask [2] = { 0xffffffff, 0xffffffff };

char  Utility [] = "HTTPD";

boolean  Debug,
         MonitorEnabled = true,
         NoSwapOut = true;

int  DclSysOutputSize = DefaultDclSysOutputSize,
     ExitStatus,
     FileBufferSize = DefaultFileBufferSize,
     NetReadBufferSize = DefaultNetReadBufferSize,
     OutputBufferSize = DefaultOutputBufferSize,
     ProcessPriority = 4,
     RejectHostsLength,
     ServerPort = 80;

unsigned long  CurrentConnectCount,
               MaxConcurrentConnections = DefaultMaxConcurrentConnections;

char  CommandLine [256],
      CommentedSoftwareID [64+24],
      SoftwareID [64];

char  *AcceptHostsPtr = NULL,
      *RejectHostsPtr = NULL;

struct  GenericListStruct  RequestList;
struct  GenericListStruct  RequestHistoryList;
int  RequestHistoryCount,
     RequestHistoryMax;

struct AnExitHandler  ExitHandler;

struct AccountingStruct  Accounting;
boolean  AccountingZeroOnStartup;

/********************/
/* external storage */
/********************/

extern boolean  AuthorizationEnabled;
extern boolean  CliLoggingDisabled;
extern boolean  CliLoggingEnabled;
extern boolean  CliMonitorDisabled;
extern boolean  CliMonitorEnabled;
extern boolean  ControlExitRequested;
extern boolean  ControlRestartRequested;
extern boolean  LoggingEnabled;
extern int  CliServerPort;
extern int  ServerHostNameLength;
extern unsigned long  ClientInternetAddress32bit;
extern unsigned short  ClientChannel;
extern unsigned short  ServerChannel;
extern char  *CliAcceptHostsPtr;
extern char  *CliRejectHostsPtr;
extern char  CliLogFileName[];
extern char  ClientHostName[];
extern char  ClientInternetAddress[];
extern char  ControlBuffer[];
extern char  ErrorRedirectionCount[];
extern char  ErrorSanityCheck[];
extern char  ErrorStringSize[];
extern char  HttpProtocol[];
extern char  ServerHostName[];
extern struct ConfigStruct  Config;

/*****************************************************************************/
/*
*/

int main ()

{
   static unsigned long  AltpriMask [2] = { PRV$M_ALTPRI, 0 },
                         PswapmMask [2] = { PRV$M_PSWAPM, 0 };

   register char  *cptr;

   int  status;
   unsigned short  Length;
   char  ProcessName [16],
         TimeScratch [32];
   $DESCRIPTOR (ProcessNameDsc, ProcessName);
   $DESCRIPTOR (ProcessNameFaoDsc, "HTTPd:!UL");

   /*********/
   /* begin */
   /*********/

   sprintf (SoftwareID, "HTTPd-WASD/%s %OpenVMS/%s %s",
            HttpdVersion, HttpdArch, HttpdIp);

   if (VMSnok (status = ParseCommandLine ()))
      exit (status);
   if (!Debug)
      if (getenv ("HTTPD$DBUG") != NULL)
         Debug = true;

   if (ControlBuffer[0])
   {
      /*********************************************************/
      /* this image is being used to control the HTTPd process */
      /*********************************************************/

      if (CliServerPort) ServerPort = CliServerPort;
      if (ServerPort < 1 || ServerPort > 65535)
         ErrorExitVmsStatus (SS$_BADPARAM, "port out of range",
                             __FILE__, __LINE__);
      exit (ControlCommand (ControlBuffer));
   }

   /****************/
   /* HTTPd server */
   /****************/

   /* output the GNU GENERAL PUBLIC LICENSE message */
   fprintf (stdout, "%%%s-I-SOFTWAREID, %s\n%s",
            Utility, SoftwareID, CopyRightMessageBrief);

   /* make sure the process' privileges are those of a mere mortal */
   status = sys$setprv (0, &ResetPrivMask, 0, 0);
   if (Debug) fprintf (stdout, "sys$setprv() %%X%08.08X\n", status);
   if (VMSnok (status))
      ErrorExitVmsStatus (status, "reseting privileges", __FILE__, __LINE__);
   status = sys$setprv (1, &AvJoePrivMask, 0, 0);
   if (Debug) fprintf (stdout, "sys$setprv() %%X%08.08X\n", status);
   if (VMSnok (status))
      ErrorExitVmsStatus (status, "default privileges", __FILE__, __LINE__);

   /* get local (server) host details */
   NetGetServerHostName ();

   /* read the configuration file */
   Configure (&Config);

   if (Config.IncludeCommentedInfo)
      sprintf (CommentedSoftwareID, "<!-- SoftwareID: %s -->\n",  SoftwareID);

   if (Config.Busy) MaxConcurrentConnections = Config.Busy;

   if (CliServerPort)
      ServerPort = CliServerPort;
   else
   if (Config.ServerPort)
      ServerPort = Config.ServerPort;
   if (ServerPort < 1 || ServerPort > 65535)
      ErrorExitVmsStatus (SS$_BADPARAM, "port out of range",
                          __FILE__, __LINE__);

   if (!CliLoggingDisabled && (Config.LoggingEnabled || CliLoggingEnabled))
      LoggingEnabled = true;
   else
      LoggingEnabled = false;

   if (!CliMonitorDisabled && (Config.MonitorEnabled || CliMonitorEnabled))
      MonitorEnabled = true;
   else
      MonitorEnabled = false;

   if (CliAcceptHostsPtr != NULL)
      Config.AcceptHostsPtr = CliAcceptHostsPtr;

   if (CliRejectHostsPtr != NULL)
      Config.RejectHostsPtr = CliRejectHostsPtr;

   /* initialize some error messages */
   ErrorInitialize ();

   /* authentication/authorization configuration */
   AuthPathConfig (NULL);

   /* set process priority */
   if (VMSnok (status = sys$setpri (0, 0, ProcessPriority, 0, 0, 0)))
      ErrorExitVmsStatus (status, "set process priority", __FILE__, __LINE__);

   /* set the process name */
   if (VMSnok (status =
       sys$fao (&ProcessNameFaoDsc, &Length, &ProcessNameDsc, ServerPort)))
      ErrorExitVmsStatus (status, "process name", __FILE__, __LINE__);
   ProcessName[ProcessNameDsc.dsc$w_length = Length] = '\0';
   if (Debug) fprintf (stdout, "ProcessName |%s|\n", ProcessName);
   if (VMSnok (status = sys$setprn (&ProcessNameDsc)))
      ErrorExitVmsStatus (status, "set process name", __FILE__, __LINE__);

   /* reset logical names provided to the HTTPd monitor utility */
   if (MonitorEnabled)
      if (VMSnok (status = DefineMonitorLogicals (NULL)))
         ErrorExitVmsStatus (status, "monitor logicals", __FILE__, __LINE__);

   /* ensure the GMT logical name is defined ok */
   if (VMSnok (status = TimeSetGmt ()))
      ErrorExitVmsStatus (status, "GMT time string", __FILE__, __LINE__);

   /* load the rule mapping database */
   MapUrl (NULL, NULL, NULL, NULL);

   /* set up to allow the HTTPD process to be controlled */
   if (VMSnok (status = ControlHttpd ()))
      ErrorExitVmsStatus (status, "control", __FILE__, __LINE__);

   /* initialize the timeout function (will exit if a problem) */
   TimeoutSet (NULL, TIMEOUT_INITIALIZE);

   /* initialize the logging */
   if (LoggingEnabled)
      if (VMSnok (status = Logging (NULL, LOGGING_BEGIN)))
         ErrorExitVmsStatus (status, "begin logging", __FILE__, __LINE__);

   /* set up and declare the exit handler */
   ExitHandler.HandlerAddress = &HttpdExit;
   ExitHandler.ArgCount = 1;
   ExitHandler.ExitStatusPtr = &ExitStatus;
   if (VMSnok (status = sys$dclexh (&ExitHandler)))
      ErrorExitVmsStatus (status, "declare exit handler", __FILE__, __LINE__);

   /* initialize DCL processing */
   DclInit ();

   /* intialize request history mechanism (ensure it's reasonable!) */
   RequestHistoryMax = Config.RequestHistory;
   if (RequestHistoryMax > 999) RequestHistoryMax = 0;
   RequestHistoryCount = 0;
   RequestHistoryList.HeadPtr = RequestHistoryList.TailPtr = NULL;

   /* set process priority */
   if (VMSnok (status = sys$setprv (1, &AltpriMask, 0, 0)))
      ErrorExitVmsStatus (status, "enabling ALTPRI", __FILE__, __LINE__);

   if (VMSnok (status = sys$setpri (0, 0, ProcessPriority, 0, 0, 0)))
      ErrorExitVmsStatus (status, "setting process priority",
                          __FILE__, __LINE__);

   if (VMSnok (status = sys$setprv (0, &AltpriMask, 0, 0)))
      ErrorExitVmsStatus (status, "disabling ALTPRI", __FILE__, __LINE__);

   /* disable process swapping */
   if (NoSwapOut)
   {
      if (VMSnok (status = sys$setprv (1, &PswapmMask, 0, 0)))
         ErrorExitVmsStatus (status, "enabling PSWAPM", __FILE__, __LINE__);

      if (VMSnok (status = sys$setswm (1)))
         ErrorExitVmsStatus (status, "setting swap mode", __FILE__, __LINE__);

      if (VMSnok (status = sys$setprv (0, &PswapmMask, 0, 0)))
         ErrorExitVmsStatus (status, "disabling PSWAPM", __FILE__, __LINE__);
   }

   /* effectively an infinite loop */
   exit (NetListen ());
}

/*****************************************************************************/
/*
Exit the HTTP server, via this declared exit handler.
*/

HttpdExit ()

{
   if (Debug) fprintf (stdout, "HttpdExit()\n %%X%08.08X", ExitStatus);

   /* run down any subprocesses associated with script processing */
   DclExit ();

   if (LoggingEnabled)
   {
      /* provide a server exit entry */
      Logging (NULL, LOGGING_END);
   }

   if (MonitorEnabled)
   {
      /* set exit status in logical names used by HTTPd monitor utility */
      Accounting.LastExitStatus = ExitStatus;
      DefineMonitorLogicals (NULL);
   }
}

/*****************************************************************************/
/*
This function can be called from two sources.  First, NetListen() where a new
connection has been established and memory has just been allocated for this
connection's thread structure.  Second, from ConcludeProcessing() where a
persistent connection (HTTP/1.0) is involved.

Both then queue a read from the network connection which calls an AST
function when the read completes, to process the request.  The keep-alive
read usually has a far shorter timeout associated with it so that rapid,
sequential reads (as happens when a document is being populated with images
for instance) are serviced from the one connection, but these are not left
open consuming resources on the off-chance of another request in the
not-too-distance future.

Persistent connections are implemented by disposing of all memory allocated
on a request's heap, and all data in the connection thread structure by
zeroing, EXCEPT that above the field 'RetainAboveZeroBelow' (client
connection information), then treating the almost-amnesic structure like a
brand-new connection (only with a shorter timeout period).
*/ 

HttpdConnection (struct RequestStruct *rqptr)

{
   int  status,
        TimerType;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "HttpdConnection()\n");

   if (!rqptr->ClientChannel)
   {
      /******************/
      /* new connection */
      /******************/

      /* add entry to the top of the request list */
      ListAddHead (&RequestList, &rqptr->RequestListEntry, rqptr);

      rqptr->ClientChannel = ClientChannel;
      rqptr->ClientInternetAddress32bit = ClientInternetAddress32bit;
      strcpy (rqptr->ClientInternetAddress, ClientInternetAddress);
      strcpy (rqptr->ClientHostName, ClientHostName);

      TimerType = TIMEOUT_INPUT;
   }
   else
   {
      /*************************/
      /* persistent connection */
      /*************************/

      memset ((unsigned char*)&rqptr->RetainAboveZeroBelow, 0,
              sizeof(struct RequestStruct) -
              ((unsigned char*)&rqptr->RetainAboveZeroBelow -
               (unsigned char*)rqptr));

      TimerType = TIMEOUT_KEEPALIVE;
   }

   if (Debug) fprintf (stdout, "KeepAliveCount: %d\n", rqptr->KeepAliveCount);

   /* initialize the per-request heap */
   rqptr->HeapHeadPtr = rqptr->HeapTailPtr = NULL;
   rqptr->HeapSize = 0;
   rqptr->AccountingDone = false;
   /* initialize the output buffer storage */
   rqptr->OutputBufferCurrentPtr = rqptr->OutputBufferPtr = NULL;
   /* timestamp the transaction */
   sys$gettim (&rqptr->BinaryTime);
   sys$numtim (&rqptr->NumericTime, &rqptr->BinaryTime);
   HttpGmTimeString (rqptr->GmDateTime, &rqptr->BinaryTime);

   /* initialize the timer for the request timeout */
   if (VMSnok (status = TimeoutSet (rqptr, TimerType)))
   {
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      return (status);
   }

   /* 
      Queue an asynchronous read to get the request from the client.
      When the read completes call RequestGet() completion function to
      further process the request.  Allow one byte for a terminating null.
   */
   if ((rqptr->NetReadBufferPtr =
       HeapAlloc (rqptr, NetReadBufferSize+1)) == NULL)
   {
      ErrorHeapAlloc (rqptr, __FILE__, __LINE__);
      return (STS$K_ERROR);
   }
   rqptr->NetReadBufferSize = NetReadBufferSize;

   if (VMSnok (status = QioNetRead (rqptr,
                                    &RequestGet,
                                    rqptr->NetReadBufferPtr,
                                    rqptr->NetReadBufferSize)))
   {
      /* new connections get an error message */
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      return (status);
   }

   return (SS$_NORMAL);
}

/****************************************************************************/
/*
The 'rqptr->LocationPtr' is non-NULL indicating redirection.  This 
can be local, indicated by a leading '/', or it can be non-local, indicated
by anything other than a leading '/', usually a full URL including scheme
(e.g. "http:").

If authorization is enabled then all redirected requests should be returned
to the client in case authorization information needs to be applied to the
new request.  This may involve reformating a local redirection into a
non-local one.  If authentication is not enabled then check both local and
non-local redirections to see if it can be handled locally.

Return normal status to indicate local-redirection (and that the thread
should NOT be disposed of), or an error status to indicate client-handled,
non-local-redirection, or a legitimate error (and that the thread can go).
*/ 

int HttpdRedirect (struct RequestStruct *rqptr)

{
   static $DESCRIPTOR (BufferDsc, "");

   static char  ContentLength [32];
   static $DESCRIPTOR (contentLengthDsc, ContentLength);
   static $DESCRIPTOR (ContentLengthFaoDsc, "Content-Length: !UL\r\n\0");

   static $DESCRIPTOR (Http302FaoDsc,
"!AZ 302 Redirection.\r\n\
Location: !AZ\r\n\
Server: !AZ\r\n\
Date: !AZ\r\n\
\r\n\0");

   static $DESCRIPTOR (LocalFaoDsc,
"!AZ !AZ !AZ\r\n\
!AZ!AZ!AZ\
!AZ!AZ!AZ\
!AZ!AZ!AZ\
!AZ!AZ!AZ\
!AZ!AZ!AZ\
!AZ!AZ!AZ\
!AZ\
!AZ\
\r\n\0");

   register unsigned long  *vecptr;
   register char  *cptr, *sptr;

   boolean  LocalRedirection;
   int  status,
        PortNumber;
   unsigned short  Length;
   unsigned long  FaoVector [32];
   char  *HttpMethodNamePtr;
   char  HttpMethodName [32],
         RedirectionUrl [1024];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "HttpdRedirect() |%s|\n", rqptr->LocationPtr);

   if (rqptr->RedirectionCount++ > 1)
   {
      rqptr->ResponseStatusCode = 501;
      ErrorGeneral (rqptr, ErrorRedirectionCount, __FILE__, __LINE__);
      return (STS$K_ERROR);
   }

   rqptr->ResponseStatusCode = 302;

   /*
      A leading space is normally not possible.
      This indicates a specific HTTP method must be used.
      Format is: "<space>METHOD<space><redirection-path>".
   */
   if (rqptr->LocationPtr[0] == ' ')
   {
      HttpMethodNamePtr = sptr = HttpMethodName;
      for (cptr = rqptr->LocationPtr + 1;
           *cptr && *cptr != ' ';
           *sptr++ = *cptr++);
      *sptr = '\0';
      rqptr->LocationPtr = cptr + 1;
   }
   else
      HttpMethodNamePtr = rqptr->HttpMethodName;

   MapUrl_VirtualPath (rqptr->PathInfoPtr, rqptr->LocationPtr,
                       RedirectionUrl, sizeof(RedirectionUrl));

   /* now ignore old value of location redirection pointer */
   rqptr->LocationPtr = NULL;

#ifdef ALWAYS_LOCAL_REDIRECTION_IF_DETECTED

   LocalRedirection = false;

   if (RedirectionUrl[0] == '/')
      LocalRedirection = true;
   else
   {
      /*****************************/
      /* check for local host/port */
      /*****************************/

      if (!memcmp (cptr = RedirectionUrl, "http:", 5))
      {
         /* scan over the "http://" */
         while (*cptr && *cptr != '/') cptr++;
         while (*cptr && *cptr == '/') cptr++;
         if (Debug) fprintf (stdout, "cptr |%s|%s|\n", cptr, ServerHostName);
         if (strsame (cptr, ServerHostName, ServerHostNameLength) &&
             (cptr[ServerHostNameLength] == '/' ||
              cptr[ServerHostNameLength] == ':'))
         {
            cptr += ServerHostNameLength;
            if (*cptr == '/')
               LocalRedirection = true;
            else
            if (*cptr == ':')
            {
               PortNumber = atoi(++cptr);
               if (Debug) fprintf (stdout, "PortNumber: %d\n", PortNumber);
               if (PortNumber == ServerPort) LocalRedirection = true;
            }
            while (*cptr && *cptr != '/') cptr++;
         }

         if (LocalRedirection)
         {
            /* eliminate host details, copy just the path over the host detail */
            sptr = RedirectionUrl;
            while (*cptr) *sptr++ = *cptr++;
            *sptr = '\0';
            if (Debug) fprintf (stdout, "RedirectUrl |%s|\n", RedirectionUrl);
         }
      }
   }

#else

   if (RedirectionUrl[0] == '/')
      LocalRedirection = true;
   else
      LocalRedirection = false;

#endif

   if (LocalRedirection)
   {
      /*********************/
      /* local redirection */
      /*********************/

      Accounting.RedirectLocalCount++;

      vecptr = &FaoVector;

      *vecptr++ = HttpMethodNamePtr;
      *vecptr++ = RedirectionUrl;
      *vecptr++ = HttpProtocol;

      if (rqptr->HttpUserAgentPtr != NULL)
      {
         *vecptr++ = "User-Agent: ";
         *vecptr++ = rqptr->HttpUserAgentPtr;
         *vecptr++ = "\r\n";
      }
      else
      {
         *vecptr++ = "";
         *vecptr++ = "";
         *vecptr++ = "";
      }

      if (rqptr->HttpRefererPtr != NULL)
      {
         *vecptr++ = "Referer: ";
         *vecptr++ = rqptr->HttpRefererPtr;
         *vecptr++ = "\r\n";
      }
      else
      {
         *vecptr++ = "";
         *vecptr++ = "";
         *vecptr++ = "";
      }

      if (rqptr->HttpAuthorizationPtr != NULL)
      {
         *vecptr++ = "Authorization: ";
         *vecptr++ = rqptr->HttpAuthorizationPtr;
         *vecptr++ = "\r\n";
      }
      else
      {
         *vecptr++ = "";
         *vecptr++ = "";
         *vecptr++ = "";
      }

      if (rqptr->HttpIfModifiedSincePtr != NULL)
      {
         *vecptr++ = "If-Modified-Since: ";
         *vecptr++ = rqptr->HttpIfModifiedSincePtr;
         *vecptr++ = "\r\n";
      }
      else
      {
         *vecptr++ = "";
         *vecptr++ = "";
         *vecptr++ = "";
      }

      if (rqptr->HttpPragmaPtr)
      {
         *vecptr++ = "Pragma: ";
         *vecptr++ = rqptr->HttpPragmaPtr;
         *vecptr++ = "\r\n";
      }
      else
      {
         *vecptr++ = "";
         *vecptr++ = "";
         *vecptr++ = "";
      }

      if (rqptr->HttpCookiePtr)
      {
         *vecptr++ = "Cookie: ";
         *vecptr++ = rqptr->HttpCookiePtr;
         *vecptr++ = "\r\n";
      }
      else
      {
         *vecptr++ = "";
         *vecptr++ = "";
         *vecptr++ = "";
      }

      if (rqptr->ContentLength)
      {
         sys$fao (&ContentLengthFaoDsc, 0, ContentLength,
                  rqptr->ContentLength);
         *vecptr++ = ContentLength;
      }
      else
         *vecptr++ = "";

      if (rqptr->KeepAliveRequest)
         *vecptr++ = "Connection: Keep-Alive\r\n";
      else
         *vecptr++ = "";

      BufferDsc.dsc$a_pointer = rqptr->NetReadBufferPtr;
      BufferDsc.dsc$w_length = rqptr->NetReadBufferSize;

      status = sys$faol (&LocalFaoDsc, &Length, &BufferDsc, &FaoVector);
      if (VMSnok (status) || status == SS$_BUFFEROVF)
      {
         rqptr->ErrorTextPtr = "sys$faol()";
         ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
         return (STS$K_ERROR);
      }

      if (Debug) fprintf (stdout, "|%s|\n", rqptr->NetReadBufferPtr);
      rqptr->RequestHeaderPtr = rqptr->NetReadBufferPtr;
      rqptr->RequestHeaderLength = Length;

      /* normal status indicates its a local redirection */
      return (SS$_NORMAL);
   }
   else
   {
      /*************************/
      /* non-local redirection */
      /*************************/

      Accounting.RedirectRemoteCount++;

      BufferDsc.dsc$a_pointer = rqptr->NetReadBufferPtr;
      BufferDsc.dsc$w_length = rqptr->NetReadBufferSize;

      status = sys$fao (&Http302FaoDsc, &Length, &BufferDsc,
                        HttpProtocol, RedirectionUrl, SoftwareID,
                        rqptr->GmDateTime);
      if (VMSnok (status) || status == SS$_BUFFEROVF)
      {
         rqptr->ErrorTextPtr = "sys$fao()";
         ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
         return (STS$K_ERROR);
      }
      if (Debug) fprintf (stdout, "|%s|\n", rqptr->NetReadBufferPtr);
      QioNetWrite (rqptr, &ConcludeProcessing,
                   rqptr->NetReadBufferPtr, Length-1);

      /* unreachable indicates it's non-local redirection :^) */
      return (SS$_UNREACHABLE);
   }
}

/*****************************************************************************/
/*
This function may be AST-delivery executed ONE OR MORE TIMES, FROM ITSELF, 
before a connection is actually disposed of.

If a next-task function was specified, and there has been no error message
generated, then execute it.

If there are any characters remaining in the output buffer then flush it with
an AST leading straight back to this function!

If an error has been generated then send it to the client once again with an
AST leading straight back to this function!

Check for redirection.  If local redirection do NOT dispose of the thread
(we'll be re-using the thread context).

If a persistent connection then treat it like a new connection and go back
to listening for the next request.

Otherwise close the client network socket, release dynamic data structures
allocated, etc.
*/

ConcludeProcessing (struct RequestStruct *rqptr)

{
   static long  Addx2 = 2;

   int  status,
        StatusCodeGroup;
   unsigned long  QuadScratch [2];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ConcludeProcessing()\n");

   /*
      Flush anything currently remaining in the thread's output buffer.
      AST back to this same function to continue disposing of the thread!
   */
   if (rqptr->OutputBufferCurrentPtr > rqptr->OutputBufferPtr)
      if (VMSok (NetBufferOutput (rqptr, &ConcludeProcessing, NULL, 0)))
         return;

   /*
      If an error has been reported send this to the client.
      AST back to this same function to continue disposing of the thread!
   */
   if (rqptr->ErrorMessagePtr != NULL)
      if (VMSok (status = ErrorSendToClient (rqptr, &ConcludeProcessing)))
         return;

   /*
      If the response header has not yet been sent then do it.
      AST back to this same function to continue disposing of the thread!
   */
   if (!rqptr->BytesTx && rqptr->ResponseHeaderPtr != NULL)
      if (VMSok (status = QioNetWrite (rqptr, &ConcludeProcessing, 0, 0)))
         return;

   if (rqptr->LocationPtr != NULL)
   {
      /***************/
      /* redirection */
      /***************/

      /* if local do NOT dispose of thread */
      if (VMSok (status = HttpdRedirect (rqptr)))
      {
         /* log the original request, before generating it redirected */
         if (LoggingEnabled) Logging (rqptr, LOGGING_ENTRY);

         /* re-call the parsing function to begin re-processing the request */
         RequestParseAndExecute (rqptr);
         return;
      }

      /*
         If a genuine error reported when redirecting send this to the client.
         AST back to this same function to continue disposing of the thread!
      */
      if (rqptr->ErrorMessagePtr != NULL)
      {
         if (VMSok (ErrorSendToClient (rqptr, &ConcludeProcessing)))
            return;
      }
      else
      {
         /*
            A status of SS$_UNREACHABLE is returned when non-local redirect.
            The network write will AST back to this same function to continue.
         */
         if (status == SS$_UNREACHABLE) return;
      }
   }

   /**************************/
   /* dispose of the REQUEST */
   /**************************/

   if (Debug) fprintf (stdout, "rqptr->BytesTx: %d\n", rqptr->BytesTx);

   /* a little more accounting */
   Accounting.ConnectCurrent = CurrentConnectCount-1;
   QuadScratch[0] = rqptr->BytesTx;
   QuadScratch[1] = 0;
   lib$addx (&QuadScratch, &Accounting.QuadBytesTx,
             &Accounting.QuadBytesTx, &Addx2);
   QuadScratch[0] = rqptr->BytesRx;
   QuadScratch[1] = 0;
   lib$addx (&QuadScratch, &Accounting.QuadBytesRx,
             &Accounting.QuadBytesRx, &Addx2);

   /* set logical names used by HTTPd monitor utility */
   if (MonitorEnabled) DefineMonitorLogicals (rqptr);

   if (!rqptr->KeepAliveTimeout)
   {
      StatusCodeGroup = rqptr->ResponseStatusCode / 100;
      if (StatusCodeGroup < 0 || StatusCodeGroup > 5) StatusCodeGroup = 0;
      Accounting.ResponseStatusCodeCountArray[StatusCodeGroup]++;

      if (LoggingEnabled) Logging (rqptr, LOGGING_ENTRY);

      /* if a request history is being kept then provide the entry */
      if (RequestHistoryMax) HttpdRequestHistory (rqptr);
   }

   /* free dynamic memory allocated for the request's heap */
   HeapFree (rqptr);

   if (rqptr->KeepAliveRequest &&
       rqptr->KeepAliveResponse &&
       rqptr->KeepAliveCount < MaxKeepAliveCount)
   {
      /*************************/
      /* PERSISTENT connection */
      /*************************/

      /*
         If the persistent connection is accepted then just return.
         Otherwise continue to completely dispose of the connection.
      */
      if (VMSok (HttpdConnection (rqptr))) return;

      /* free any dynamic memory that may have been reallocated to the heap */
      HeapFree (rqptr);
   }

   /*****************************/
   /* dispose of the CONNECTION */
   /*****************************/

   /* remove from the request list */
   ListRemove (&RequestList, &rqptr->RequestListEntry);

   NetShutdownSocket (rqptr->ClientChannel);

   /* cancel any outstanding timeout timer */
   sys$cantim (rqptr, 0);

   /* free dynamic memory allocated for the connection structure */
   if (Debug) fprintf (stdout, "free() rqptr\n");
   free (rqptr);

   /* one less to worry about! */
   if (CurrentConnectCount) CurrentConnectCount--;

   /* don't bother checking for exit or restart if connections still exist */
   if (CurrentConnectCount) return;

   /* if the control module has requested server exit or restart */
   if (ControlExitRequested)
   {
      fprintf (stdout, "%%%s-I-CONTROL, delayed server exit\n", Utility);
      ExitStatus = SS$_NORMAL;
      HttpdExit();
      sys$delprc (0, 0);
   }
   else
   if (ControlRestartRequested)
   {
      fprintf (stdout, "%%%s-I-CONTROL, delayed server restart\n", Utility);
      exit (SS$_NORMAL);
   }
}

/*****************************************************************************/
/*
Keeps a linked list history from most to least recent that can be reported on
by AdminRequestReport().
*/

int HttpdRequestHistory (struct RequestStruct *rqptr)

{
   register char  *cptr, *sptr, *zptr;
   register struct RequestHistoryStruct  *hlptr;

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "HttpdRequestHistory()\n");

   if (RequestHistoryCount < RequestHistoryMax)
   {
      /* allocate memory for a new entry */
      if ((hlptr = calloc (1, sizeof (struct RequestHistoryStruct))) == NULL)
         ErrorExitVmsStatus (status, "calloc()", __FILE__, __LINE__);
      RequestHistoryCount++;
   }
   else
   {
      /* reuse the tail entry (least recent) */
      hlptr = RequestHistoryList.TailPtr;
      ListRemove (&RequestHistoryList, &hlptr->HistoryListEntry);
   }

   /* add entry to the head of the history list (most recent) */
   ListAddHead (&RequestHistoryList, &hlptr->HistoryListEntry, hlptr);

   memcpy (&hlptr->BinaryTime, &rqptr->BinaryTime, 8);
   hlptr->BytesRx = rqptr->BytesRx;
   hlptr->BytesTx = rqptr->BytesTx;
   hlptr->ResponseStatusCode = rqptr->ResponseStatusCode;

   zptr = (sptr = hlptr->ClientAndRequest) + sizeof(hlptr->ClientAndRequest);

   if (rqptr->RemoteUser[0])
   {
      cptr = rqptr->RemoteUser;
      while (*cptr && sptr < zptr) *sptr++ = *cptr++;
      if (sptr < zptr) *sptr++ = '@';
   }
   cptr = rqptr->ClientHostName;
   while (*cptr && sptr < zptr) *sptr++ = *cptr++;
   if (sptr < zptr) *sptr++ = '\0';
   hlptr->RequestPtr = sptr;

   if (rqptr->PathInfoPtr != NULL && rqptr->QueryStringPtr != NULL)
   {
      cptr = rqptr->HttpMethodName;
      while (*cptr && sptr < zptr) *sptr++ = *cptr++;
      if (sptr < zptr) *sptr++ = ' ';
      cptr = rqptr->ScriptName;
      while (*cptr && sptr < zptr) *sptr++ = *cptr++;
      cptr = rqptr->PathInfoPtr;
      while (*cptr && sptr < zptr) *sptr++ = *cptr++;
      if (rqptr->QueryStringPtr[0] && sptr < zptr) *sptr++ = '?';
      cptr = rqptr->QueryStringPtr;
      while (*cptr && sptr < zptr) *sptr++ = *cptr++;
   }

   if (sptr >= zptr)
   {
      hlptr->Truncated = true;
      sptr--;
   }
   else
      hlptr->Truncated = false;
   *sptr = '\0';
}

/*****************************************************************************/
/*
Input, output and "keep-alive" timeout timers.  Will exit if the initialization
encounters an error.  Cancels any existing timer before initiating another,
only one timer should be active at any one time.  Outstanding timers will be
cancelled by ConcludeProcessing() as a request is finished.
*/

int TimeoutSet
(
struct RequestStruct *rqptr,
int Function
)
{
   static unsigned long  InputTimeoutDeltaBinTime [2],
                         KeepAliveTimeoutDeltaBinTime [2],
                         OutputTimeoutDeltaBinTime [2];

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TimeoutSet() %d\n", Function);

   switch (Function)
   {
      case TIMEOUT_INITIALIZE :
      {
         /* local storage */
         unsigned short  Length;
         char  InputTimeout [32],
               KeepAliveTimeout [32],
               OutputTimeout [32];
         $DESCRIPTOR (InputTimeoutDsc, InputTimeout);
         $DESCRIPTOR (KeepAliveTimeoutDsc, KeepAliveTimeout);
         $DESCRIPTOR (KeepAliveTimeoutFaoDsc, "0 00:00:!UL.00");
         $DESCRIPTOR (OutputTimeoutDsc, OutputTimeout);
         $DESCRIPTOR (TimeoutFaoDsc, "0 00:!UL:00.00");

         /* INPUT timeout */
         if (!Config.InputTimeoutMinutes)
            Config.InputTimeoutMinutes = DefaultInputTimeoutMinutes;
         if (VMSnok (status =
             sys$fao (&TimeoutFaoDsc, &Length, &InputTimeoutDsc,
                      Config.InputTimeoutMinutes)))
            ErrorExitVmsStatus (status, "input timeout", __FILE__, __LINE__);
         InputTimeout[InputTimeoutDsc.dsc$w_length = Length] = '\0';
         if (Debug) fprintf (stdout, "InputTimeout |%s|\n", InputTimeout);
         if (VMSnok (status =
             sys$bintim (&InputTimeoutDsc, &InputTimeoutDeltaBinTime)))
            ErrorExitVmsStatus (status, "input timeout", __FILE__, __LINE__);

         /* OUTPUT timeout */
         if (!Config.OutputTimeoutMinutes)
            Config.OutputTimeoutMinutes = DefaultOutputTimeoutMinutes;
         if (VMSnok (status =
             sys$fao (&TimeoutFaoDsc, &Length, &OutputTimeoutDsc,
                      Config.OutputTimeoutMinutes)))
            ErrorExitVmsStatus (status, "output timeout", __FILE__, __LINE__);
         OutputTimeout[OutputTimeoutDsc.dsc$w_length = Length] = '\0';
         if (Debug) fprintf (stdout, "OutputTimeout |%s|\n", OutputTimeout);
         if (VMSnok (status =
             sys$bintim (&OutputTimeoutDsc, &OutputTimeoutDeltaBinTime)))
            ErrorExitVmsStatus (status, "output timeout", __FILE__, __LINE__);

         /* KEEPALIVE timeout */
         if (Config.KeepAliveTimeoutSeconds)
            status = sys$fao (&KeepAliveTimeoutFaoDsc, &Length,
                              &KeepAliveTimeoutDsc,
                              Config.KeepAliveTimeoutSeconds);
         else
            status = sys$fao (&KeepAliveTimeoutFaoDsc, &Length,
                              &KeepAliveTimeoutDsc,
                              DefaultKeepAliveTimeoutSeconds);
         if (VMSnok (status))
            ErrorExitVmsStatus (status, "output timeout", __FILE__, __LINE__);
         KeepAliveTimeout[KeepAliveTimeoutDsc.dsc$w_length = Length] = '\0';
         if (Debug)
            fprintf (stdout, "KeepAliveTimeout |%s|\n", KeepAliveTimeout);
         if (VMSnok (status =
             sys$bintim (&KeepAliveTimeoutDsc, &KeepAliveTimeoutDeltaBinTime)))
            ErrorExitVmsStatus (status, "output timeout", __FILE__, __LINE__);

         return (SS$_NORMAL);
      }

      case TIMEOUT_INPUT :

         sys$cantim (rqptr, 0);
         return (sys$setimr (0, &InputTimeoutDeltaBinTime,
                             &TimeoutAst, rqptr, 0));

      case TIMEOUT_OUTPUT :

         sys$cantim (rqptr, 0);
         return (sys$setimr (0, &OutputTimeoutDeltaBinTime,
                             &TimeoutAst, rqptr, 0));

      case TIMEOUT_KEEPALIVE :

         sys$cantim (rqptr, 0);
         return (sys$setimr (0, &KeepAliveTimeoutDeltaBinTime,
                             &TimeoutAst, rqptr, 0));

      default :
         return (SS$_BADPARAM);
   }
}

/*****************************************************************************/
/*
This AST is called when the sys$setimr() expires.  This is done to cleanup 
requests that "hang", often due to a client aborting a data transfer, or a 
script getting into an infinite loop, etc.
*/ 

TimeoutAst (struct RequestStruct *rqptr)

{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TimeoutAst()\n");

   if (rqptr->DclTaskPtr != NULL)
   {
      /* DCL involved, do any DCL-specific cleanup */
      DclConcludeTask (rqptr->DclTaskPtr, true);
   }
   else
   {
      /* cancel network I/O, should force the connection to abort */
      sys$cancel (rqptr->ClientChannel);
   }
}

/****************************************************************************/

