/*****************************************************************************/
/*
                                  hdisk.c

This application links to the Gift.c application's GIF and graphic code!
Functions begining Gift...() and storage beginning Graphic... is Gift.c code.

HyperDisk!  Provides disk details, including a usage bar graph, for one or
more disks specified as part of the path or as a CGI form variable "DISKS=".
Disk device names may be specified separated by commas, forward-slashes, or
(probably most simply) just with a full-colon (which is used by VMS as a
device name separator).  Disk details include blocks and Mbytes total, used
and free, as well as error count.  Component devices of volumes sets are
displayed individually.

Variations allowed in specifying disks:

   http://host/disk/device
   http://host/disk/device1:device2:device3
   http://host/disk/device1,device2,device3
   http://host/disk/device1/device2/device3
   http://host/disk?disk=device
   http://host/disk?disk=device1:device2:device3
   http://host/disk?disk=device1,device2,device3
   http://host/disk?disk=device1/device2/device3

If the string "graphiconly" occurs in the query string only a bar-graph GIF
for the specified device is generated and sent.  This allows documents to
include a tag such as <IMG SRC="/disk/sys$sysdevice?graphiconly"> to include
the usage bar-graph for any given device (of course only one may be specified
with each use).

Uses the Netscape HTML extension '<FONT COLOR="#nnnnnn">...</FONT>' to add some
highlighting to problem condition text.


ACKNOWLEDGEMENT
---------------
Although the graphic generation and GIF header code is mine the actual
GIF compression code employed is derived from the PBM suite, is copyright,
and used within the owner's guidelines:

 * GIF Image compression - LZW algorithm implemented with Trie type
 *                         structure.
 *                         Written by Bailey Brown, Jr.
 *                         last change May 24, 1990
 *                         file: compgif.c
 *
 *  You may use or modify this code as you wish, as long as you mention
 *  my name in your documentation.
 *
 *                  - Bailey Brown, Jr.


CGI VARIABLES
-------------
WWW_FORM_BRIEF          if the word brief occurs in the query string only
                        the bar graph is generated, no device details
WWW_FORM_DISK           one of more disk device names
WWW_FORM_DISKS          ditto
WWW_FORM_HEIGHT         height in pixels of the bar-graph
WWW_FORM_GRAPHICONLY    if the word graphiconly occurs in the query string a
                        gif representing the device is generated and returned
WWW_FORM_HELP           if the word help occurs in the query string it
                        results in a redirection to the help document
WWW_FORM_PERCENTAGE     results in a bar graph GIF being returned
WWW_FORM_WIDTH          width in pixels of the bar-graph
WWW_SCRIPT_NAME         script name
WWW_SERVER_NAME         server host name
WWW_SERVER_PORT         server IP port
WWW_QUERY_STRING        query string
WWW_PATH_INFO           document path (everything following the script name)


BUILD DETAILS
-------------
See BUILD_HDISK.COM procedure.


VERSION HISTORY (update SoftwareID as well!)
---------------
18-APR-97  MGD  v1.1.0, graphics pre-expired (generated each time accessed)
06-AUG-96  MGD  v1.0.0, YAQH (wanted to try generating coloured GIFs :^)
*/
/*****************************************************************************/

#ifdef __ALPHA
   char SoftwareID [] = "HDISK AXP-1.1.0";
#else
   char SoftwareID [] = "HDISK VAX-1.1.0";
#endif

/* standard C header files */
#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <file.h>

/* VMS header files */
#include <dcdef.h>
#include <devdef.h>
#include <descrip.h>
#include <dvidef.h>
#include <libdef.h>
#include <libdtdef.h>
#include <ssdef.h>
#include <stsdef.h>

/* application header file */
#include "[-.gift]gift.h"

/*
Can't find any UCB definitions anywhere.  Cobbled these up out of the $GETDVI 
system service description of DVI$_STS item and the VMS Device Support 
Reference describing Unit Control Blocks. 
*/ 
#define UCB$V_TIM 0x1
#define UCB$V_INT 0x2
#define UCB$V_ERLOGIP 0x4
#define UCB$V_CANCEL 0x8
#define UCB$V_ONLINE 0x10
#define UCB$V_POWER 0x20
#define UCB$V_TIMOUT 0x40
#define UCB$V_INTTYPE 0x80
#define UCB$V_BSY 0x100
#define UCB$V_MOUNTING 0x200
#define UCB$V_DEADMO 0x400
#define UCB$V_VALID 0x800
#define UCB$V_UNLOAD 0x1000
#define UCB$V_TEMPLATE 0x2000
#define UCB$V_MNTVERIP 0x4000
#define UCB$V_WRONGVOL 0x8000
#define UCB$V_DELETEUCB 0x10000
#define UCB$V_LCL_VALID 0x20000
#define UCB$V_SUPMVMSG 0x40000
#define UCB$V_MNTVERPND 0x80000
#define UCB$V_DISMOUNT 0x100000
#define UCB$V_CLUTRAN 0x200000
#define UCB$V_WRTLOCKMV 0x400000
#define UCB$V_SVPN_END 0x800000

#define VMSok(x) ((x) & STS$M_SUCCESS)
#define VMSnok(x) !(((x) & STS$M_SUCCESS))

#define boolean int
#define true 1
#define false 0

/* macro provides NULL pointer if CGI variable does not exist */
#define GetCgiVarIfExists(CharPointer,CgiVariableName) \
   CharPointer = getenv(CgiVariableName)

/* macro provides pointer to empty string even if CGI variable does not exist */
#define GetCgiVar(CharPointer,CgiVariableName) \
   if ((CharPointer = getenv(CgiVariableName)) == NULL) \
       CharPointer = ""; \
   if (Debug) fprintf (stdout, "%s |%s|\n", CgiVariableName, CharPointer);

#define ExcessiveErrorCount 5

char  Http200Header [] =
"HTTP/1.0 200 Document follows.\r\n\
Content-Type: text/html\r\n\
Expires: Thu, 1-Jan-1970 00:00:00 GMT\r\n\
\r\n";

char  Http302Header [] =
"HTTP/1.0 302 Redirection.\r\n\
Location: http://%s:%s%s\r\n\
\r\n";

char  Http404Header [] =
"HTTP/1.0 404 Error report follows.\r\n\
Content-Type: text/html\r\n\
\r\n";

char  HelpUrl [] = "/hdisk/-/hdisk.html",
      IconImg [] = "<IMG SRC=\"/hdisk/-/hdisk.gif\" ALT=\"HyperDisk\">";

struct DiskInfoStruct
{
   struct DiskInfoStruct  *NextPtr;
   long  DevClass,
         DevChar,
         DevChar2,
         ErrCnt,
         FreeBlocks,
         MaxBlock,
         UsedBlocks,
         VolNumber,
         VolSetMem;
   char  FullDevName [65],
         NextDevName [65],
         MediaName [65];
   struct {
      unsigned long  lo32;
      unsigned long  hi32;
   } UcbStatus;
};

boolean  Debug,
         DoDiskInfoTable,
         DoGraphicOnly,
         HttpHasBeenOutput;

unsigned long  DayOfWeek;
unsigned long  CurrentBinTime [2];
unsigned short  CurrentNumTime [7];
char  DayDateTime [64];

char  *CgiFormDiskPtr,
      *CgiFormHeightPtr,
      *CgiFormPercentagePtr,
      *CgiFormWidthPtr,
      *CgiPathInfoPtr,
      *CgiQueryStringPtr,
      *CgiScriptNamePtr,
      *CgiServerNamePtr,
      *CgiServerPortPtr;

FILE  *HttpOut;

char  *MonthName [] =
{ "", "January", "February", "March", "April", "May", "June",
      "July", "August", "September", "October", "November", "December" };

char  *DayName [] =
{ "", "Monday", "Tuesday", "Wednesday", "Thursday",
      "Friday", "Saturday", "Sunday" };

/* storage in the Gift.c module */
extern boolean  GraphicNoCacheGif;
extern int  GraphicHeight,
            GraphicWidth;
extern struct GraphicFontStruct Font [];

/*****************************************************************************/
/*
*/

main ()

{
   register char  *cptr, *sptr, *zptr;

   int  status,
        Percentage;
   char  UpCaseDiskName [256];

   /*********/
   /* begin */
   /*********/

   if (getenv ("HDISK$DBUG") != NULL)
   {
      Debug = true;
      system ("SHOW SYM WWW_*");
   }

   /* open another output stream so that the "\n" are not filtered */
#ifdef __DECC
   if ((HttpOut = fopen ("SYS$OUTPUT", "w", "ctx=bin")) == NULL)
      exit (vaxc$errno);
#else
   if ((HttpOut = fopen ("SYS$OUTPUT", "w", "rfm=udf")) == NULL)
      exit (vaxc$errno);
#endif

   GetCgiVar (CgiScriptNamePtr, "WWW_SCRIPT_NAME");
   GetCgiVar (CgiPathInfoPtr, "WWW_PATH_INFO");
   GetCgiVar (CgiQueryStringPtr, "WWW_QUERY_STRING");

   if (strstr (CgiQueryStringPtr, "hyperdisk") != NULL ||
       strstr (CgiQueryStringPtr, "HYPERDISK") != NULL)
   {
      GiftButton ("", "HyperDisk", 14);
      exit (SS$_NORMAL);
   }

   if (strstr (CgiQueryStringPtr, "furtherinfo") != NULL ||
       strstr (CgiQueryStringPtr, "FURTHERINFO") != NULL)
   {
      GiftButton ("", "  further\ninformation", 8);
      exit (SS$_NORMAL);
   }

   if (strstr (CgiQueryStringPtr, "brief") != NULL ||
       strstr (CgiQueryStringPtr, "BRIEF") != NULL)
      DoDiskInfoTable = false;
   else
      DoDiskInfoTable = true;

   if (strstr (CgiQueryStringPtr, "graphiconly") != NULL ||
       strstr (CgiQueryStringPtr, "GRAPHICONLY") != NULL)
      DoGraphicOnly = GraphicNoCacheGif = true;
   else
      DoGraphicOnly = GraphicNoCacheGif = false;

   GetCgiVar (CgiFormWidthPtr, "WWW_FORM_WIDTH");
   if (CgiFormWidthPtr[0])
   {
      GraphicWidth = atol(CgiFormWidthPtr);
      if (GraphicWidth <= 0 || GraphicWidth >= MaxGraphicWidth)
         exit (ErrorGeneral ("Graphic width error.",
                             __FILE__, __LINE__));
   }

   GetCgiVar (CgiFormHeightPtr, "WWW_FORM_HEIGHT");
   if (CgiFormHeightPtr[0])
   {
      GraphicHeight = atol(CgiFormHeightPtr);
      if (GraphicHeight <= 0 || GraphicHeight >= MaxGraphicHeight)
         exit (ErrorGeneral ("Graphic height error.",
                             __FILE__, __LINE__));
   }

   GetCgiVar (CgiFormPercentagePtr, "WWW_FORM_PERCENTAGE");
   if (CgiFormPercentagePtr[0])
   {
      /* provide a bar graph */
      Percentage = atol (CgiFormPercentagePtr);
      if (!isdigit(CgiFormPercentagePtr[0]) || Percentage > 100)
         exit (ErrorGeneral ("Percentage is not acceptable.",
                             __FILE__, __LINE__));
      GiftUsageBar (Percentage, 85, "");
      exit (SS$_NORMAL);
   }

   if (strlen(CgiPathInfoPtr) > 1)
      cptr = CgiPathInfoPtr+1;
   else
   {
      /* disks not in path, see if they are specified from a form */
      GetCgiVar (CgiFormDiskPtr, "WWW_FORM_DISK");
      /* nothing in singular "_disk", try plural "_disks" */
      if (!CgiFormDiskPtr[0]) GetCgiVar (CgiFormDiskPtr, "WWW_FORM_DISKS");
      /* no, no disks at all, provide some help! */
      if (!CgiFormDiskPtr[0]) exit (RedirectToHelp ());
      cptr = CgiFormDiskPtr;
   }

   zptr = (sptr = UpCaseDiskName) + sizeof(UpCaseDiskName);
   for (/* see above*/; *cptr && sptr < zptr; *sptr++ = toupper(*cptr++));
   if (sptr >= zptr)
      exit (ErrorGeneral ("String too small.", __FILE__, __LINE__));
   *sptr = '\0';

   if (DoGraphicOnly)
   {
      OneDiskInfo (UpCaseDiskName);
      exit (SS$_NORMAL);
   }

   sys$gettim (&CurrentBinTime);
   sys$numtim (&CurrentNumTime, &CurrentBinTime);

   if (VMSnok (status = lib$day_of_week (&CurrentBinTime, &DayOfWeek)))
      exit (ErrorVmsStatus (status, "day of week", 0, __FILE__, __LINE__));
   sprintf (DayDateTime, "%s, %d %s %d %02.02d:%02.02d:%02.02d",
            DayName[DayOfWeek], CurrentNumTime[2],
            MonthName[CurrentNumTime[1]], CurrentNumTime[0],
            CurrentNumTime[3], CurrentNumTime[4], CurrentNumTime[5]);

   ReportDiskInfo (UpCaseDiskName);

   exit (SS$_NORMAL);
}

/*****************************************************************************/
/*
*/

int RedirectToHelp ()

{
   GetCgiVar (CgiServerNamePtr, "WWW_SERVER_NAME");
   GetCgiVar (CgiServerPortPtr, "WWW_SERVER_PORT");
   fprintf (HttpOut, Http302Header,
            CgiServerNamePtr, CgiServerPortPtr, HelpUrl);
   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
*/

int ReportDiskInfo (char *DiskName)

{
   register char  *cptr, *sptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ProvideDiskInfo()\n");

   fprintf (HttpOut,
"%s\
<!-- SoftwareID: %s -->\n\
<HTML>\n\
<TITLE>HyperDisk: %s</TITLE>\n\
<A HREF=\"%s\">%s</A> <I>at %s</I>\n\
<P><HR SIZE=3>\n",
   Http200Header, SoftwareID,
   DiskName, HelpUrl, IconImg, DayDateTime);
   HttpHasBeenOutput = true;

   cptr = DiskName;
   while (*cptr)
   {
      sptr = cptr;
      while (*cptr && isspace(*cptr)) cptr++;
      while (*cptr && *cptr != ':' && *cptr != ',' && *cptr != '/') cptr++;
      if (*cptr) *cptr++ = '\0';
      if (!*sptr) continue;
      OneDiskInfo (sptr);
   }

   fprintf (HttpOut, "<P><HR SIZE=3>\n</HTML>\n");
}

/*****************************************************************************/
/*
*/

int OneDiskInfo (char *DiskName)

{
   int  status,
        ErrorCount,
        VolumeCount;
   unsigned long  PercentageUsed,
                  TotalMaxBlock,
                  TotalFreeBlocks,
                  TotalUsedBlocks;
   struct DiskInfoStruct  *DiskInfoPtr,
                          *FirstDiskInfoPtr,
                          *NextDiskInfoPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "OneDiskInfo()\n");

   if ((FirstDiskInfoPtr = calloc (sizeof(struct DiskInfoStruct), 1)) == NULL)
      exit (ErrorVmsStatus (vaxc$errno, "allocating memory", "",
                            __FILE__, __LINE__));

   if (VMSnok (status = GetDiskInfo (DiskName, FirstDiskInfoPtr)))
   {
      free (FirstDiskInfoPtr);
      if (status == SS$_NOSUCHDEV)
      {
         fprintf (HttpOut,
            "<H2>%s ... <FONT COLOR=\"#ff0000\">NO SUCH DEVICE!</FONT></H2>\n",
             DiskName);
         return;
      }
      exit (ErrorVmsStatus (status, DiskName, "", __FILE__, __LINE__));
   }

   if (!(FirstDiskInfoPtr->DevClass & DC$_DISK))
   {
      free (FirstDiskInfoPtr);
      fprintf (HttpOut,
         "<H2>%s ... <FONT COLOR=\"#ff0000\">NOT A DISK DEVICE!</FONT></H2>\n",
         DiskName);
      return;
   }

   if (!(FirstDiskInfoPtr->DevChar & DEV$M_MNT))
   {
      free (FirstDiskInfoPtr);
      fprintf (HttpOut,
         "<H2>%s ... <FONT COLOR=\"#ff0000\">NOT MOUNTED!</FONT></H2>\n",
         DiskName);
      return;
   }

   /* on-line bit should normally be 1 */
   if (!(FirstDiskInfoPtr->UcbStatus.lo32 & UCB$V_ONLINE))
   {
      free (FirstDiskInfoPtr);
      fprintf (HttpOut,
         "<H2>%s ... <FONT COLOR=\"#ff0000\">NOT ONLINE!</FONT></H2>\n",
         DiskName);
      return;
   }

   /* unit timeout bit should normally be 0 */
   if (FirstDiskInfoPtr->UcbStatus.lo32 & UCB$V_TIMOUT)
   {
      free (FirstDiskInfoPtr);
      fprintf (HttpOut,
"<H2>%s ... <FONT COLOR=\"#ff0000\">MOUNT VERIFY HAS TIMED-OUT!</FONT></H2>\n\
<P>This device was in a mount verify state, <B>but</B> that has now \
timed-out.  It could have been off-line, had the wrong volume on-line, \
had excessive errors or a corrupted file system.  <B>In short ... \
there is a major problem that should investigated ASAP.</B>\n",
               DiskName);
      return;
   }

   /* mount verify bit should normally be 0 */
   if (FirstDiskInfoPtr->UcbStatus.lo32 & UCB$V_MNTVERIP)
   {
      free (FirstDiskInfoPtr);
      fprintf (HttpOut,
"<H2>%s ... <FONT COLOR=\"#ff0000\">IN MOUNT VERIFY!</FONT></H2>\n\
<P>This device is in a mount verify state.  It could be off-line, \
have the wrong volume on-line, have excessive errors or a corrupted \
file system.  <B>In short ... there is a major problem that should \
investigated ASAP.</B>\n",
               DiskName);
      return;
   }

   if (FirstDiskInfoPtr->VolSetMem && FirstDiskInfoPtr->VolNumber > 1)
   {
      free (FirstDiskInfoPtr);
      fprintf (HttpOut,
"<H2>%s ... <FONT COLOR=\"#ff0000\">NOT PRIMARY MEMBER!</FONT></H2>\n\
<P>This device is part of a volume set, but is not the primary member \
of that set.  Specify only the primary member!\n",
               DiskName);
      return;
   }

   /***************************************/
   /* additional devices for a volume set */
   /***************************************/

   DiskInfoPtr = FirstDiskInfoPtr;
   while (DiskInfoPtr->NextDevName[0])
   {
      if ((DiskInfoPtr->NextPtr =
           calloc (sizeof(struct DiskInfoStruct), 1)) == NULL)
         exit (ErrorVmsStatus (vaxc$errno, "allocating memory", "",
                               __FILE__, __LINE__));

      GetDiskInfo (DiskInfoPtr->NextDevName, DiskInfoPtr->NextPtr);
      DiskInfoPtr = DiskInfoPtr->NextPtr;
   }

   /********************/
   /* calculate totals */
   /********************/

   ErrorCount = TotalMaxBlock = TotalFreeBlocks = TotalUsedBlocks =
      VolumeCount = 0;
   DiskInfoPtr = FirstDiskInfoPtr;
   while (DiskInfoPtr != NULL)
   {
      VolumeCount++;
      ErrorCount += DiskInfoPtr->ErrCnt;
      TotalMaxBlock += DiskInfoPtr->MaxBlock;
      TotalFreeBlocks += DiskInfoPtr->FreeBlocks;
      TotalUsedBlocks += DiskInfoPtr->UsedBlocks;
      DiskInfoPtr = DiskInfoPtr->NextPtr;
   }

   /* round up or down? */
   if (((TotalUsedBlocks * 10) / (TotalMaxBlock / 100)) % 10 >= 5)
      PercentageUsed = ((TotalUsedBlocks * 10) / (TotalMaxBlock / 10)) + 1;
   else
      PercentageUsed = (TotalUsedBlocks * 10) / (TotalMaxBlock / 10);

   /****************/
   /* graphic only */
   /****************/

   if (DoGraphicOnly)
   {
      /* local storage */
      char  String [256];

      if (PercentageUsed == 100)
         sprintf (String, "%s  %d%% used", DiskName, PercentageUsed);
      else
      if (PercentageUsed)
         sprintf (String, "%s  %d%% used %d%% free",
                  DiskName, PercentageUsed, 100 - PercentageUsed);
      else
         sprintf (String, "%s %d%% free", DiskName, 100 - PercentageUsed);
      GiftUsageBar (PercentageUsed, 85, String);
      /* no need to deallocate memory, will be exit()ing shortly */
      return;
   }

   /*********************/
   /* graphic image tag */
   /*********************/

   fprintf (HttpOut,
"<H2>%s ... used: %d%% free: %d%%",
            DiskName, PercentageUsed, 100 - PercentageUsed);

   if (ErrorCount >= ExcessiveErrorCount)
      fprintf (HttpOut, " <FONT COLOR=\"#ff0000\">ERRORS!</FONT>");

   fprintf (HttpOut,
"</H2>\n\
<IMG SRC=\"%s?percentage=%d\" ALT=\"[disk usage indicator]\">\n",
            CgiScriptNamePtr, PercentageUsed);

   if (DoDiskInfoTable)
   {
      /***************************************/
      /* present table of device information */
      /***************************************/

      if (VolumeCount == 1)
         fprintf (HttpOut, "<P>");
      else
         fprintf (HttpOut, "<P>Volume set comprising %d devices.\n<BR>",
                  VolumeCount);

      fprintf (HttpOut,
"<TABLE BORDER>\n\
<TH>\
<TD COLSPAN=2>Total <TT>(blk|Mb)</TT></TD>\
<TD COLSPAN=3>Used <TT>(blk|Mb|%%)</TT></TD>\
<TD COLSPAN=3>Free <TT>(blk|Mb|%%)</TT></TD>\
<TD>Errors</TD>\
<TD>Media</TD>\
</TH>\n");

      /* all disks (just one, or multiple if a volume set) */

      DiskInfoPtr = FirstDiskInfoPtr;
      while (DiskInfoPtr != NULL)
      {
         PercentageUsed = (DiskInfoPtr->UsedBlocks * 100) /
                          DiskInfoPtr->MaxBlock;

         fprintf (HttpOut,
"<TR ALIGN=RIGHT>\
<TD ALIGN=LEFT><TT>%s</TT></TD>\
<TD>%d</TD>\
<TD>%d</TD>\
<TD>%d</TD>\
<TD>%d</TD>\
<TD>%d%%</TD>\
<TD>%d</TD>\
<TD>%d</TD>\
<TD>%d%%</TD>\
<TD>%d</TD>\
<TD>%s</TD>\
</TR>\n",
         DiskInfoPtr->FullDevName,
         DiskInfoPtr->MaxBlock,
         DiskInfoPtr->MaxBlock >> 11,
         DiskInfoPtr->UsedBlocks,
         DiskInfoPtr->UsedBlocks >> 11,
         PercentageUsed,
         DiskInfoPtr->FreeBlocks,
         DiskInfoPtr->FreeBlocks >> 11,
         100 - PercentageUsed,
         DiskInfoPtr->ErrCnt,
         DiskInfoPtr->MediaName);

         DiskInfoPtr = DiskInfoPtr->NextPtr;
      }

      /* block and Mbyte totals */

      fprintf (HttpOut,
"<TR ALIGN=RIGHT>\
<TD></TD>\
<TD>%d</TD>\
<TD></TD>\
<TD>%d</TD>\
<TD></TD>\
<TD></TD>\
<TD>%d</TD>\
<TD></TD>\
<TD></TD>\
<TD></TD>\
<TD></TD>\
</TR>\n\
<TR ALIGN=RIGHT>\
<TD></TD>\
<TD></TD>\
<TD>%d</TD>\
<TD></TD>\
<TD>%d</TD>\
<TD></TD>\
<TD></TD>\
<TD>%d</TD>\
<TD></TD>\
<TD></TD>\
<TD></TD>\
</TR>\n",
      TotalMaxBlock,
      TotalUsedBlocks,
      TotalFreeBlocks,
      TotalMaxBlock >> 11,
      TotalUsedBlocks >> 11,
      TotalFreeBlocks >> 11);

      fprintf (HttpOut, "</TABLE>\n");
   }

   /*************************/
   /* free allocated memory */
   /*************************/

   NextDiskInfoPtr = FirstDiskInfoPtr;
   while (NextDiskInfoPtr != NULL)
   {
      DiskInfoPtr = NextDiskInfoPtr;
      NextDiskInfoPtr = NextDiskInfoPtr->NextPtr;
      free (DiskInfoPtr);
   }
}

/*****************************************************************************/
/*
*/

GetDiskInfo
(
char *DiskName,
struct DiskInfoStruct *diptr
)
{
   static long  DviDevClass = DVI$_DEVCLASS,
                DviDevChar = DVI$_DEVCHAR,
                DviDevChar2 = DVI$_DEVCHAR2,
                DviFullDevNam = DVI$_FULLDEVNAM,
                DviFreeBlocks = DVI$_FREEBLOCKS,
                DviErrCnt = DVI$_ERRCNT,
                DviMaxBlock = DVI$_MAXBLOCK,
                DviMediaName = DVI$_MEDIA_NAME,
                DviNextDevNam = DVI$_NEXTDEVNAM,
                DviSts = DVI$_STS,
                DviVolNumber = DVI$_VOLNUMBER,
                DviVolSetMem = DVI$_VOLSETMEM;

   register char  *cptr, *sptr;

   int  status;
   unsigned short  StringItemLength;
   $DESCRIPTOR (DeviceNameDsc, "");
   $DESCRIPTOR (StringItemDsc, "");

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "GetDiskInfo() |%s|\n", DiskName);

   diptr->MaxBlock = diptr->FreeBlocks = diptr->UsedBlocks = 0;
   diptr->FullDevName[0] = diptr->NextDevName[0] = diptr->MediaName[0] = '\0';
   diptr->UcbStatus.lo32 = diptr->UcbStatus.hi32 = 0;

   DeviceNameDsc.dsc$w_length = strlen (DiskName);
   DeviceNameDsc.dsc$a_pointer = DiskName;

   if (VMSnok (status =
       lib$getdvi (&DviDevClass, 0, &DeviceNameDsc, &diptr->DevClass, 0, 0)))
      return (status);

   if (!(diptr->DevClass & DC$_DISK)) return (SS$_NORMAL);

   if (VMSnok (status =
       lib$getdvi (&DviDevChar, 0, &DeviceNameDsc, &diptr->DevChar, 0, 0)))
      return (status);

   if (!(diptr->DevChar & DEV$M_MNT)) return (SS$_NORMAL);

   if (VMSnok (status =
       lib$getdvi (&DviDevChar2, 0, &DeviceNameDsc, &diptr->DevChar2, 0, 0)))
      return (status);

   if (VMSnok (status =
       lib$getdvi (&DviSts, 0, &DeviceNameDsc, &diptr->UcbStatus, 0, 0)))
      return (status);

   StringItemDsc.dsc$w_length = sizeof(diptr->FullDevName);
   StringItemDsc.dsc$a_pointer = diptr->FullDevName;

   if (VMSnok (status =
       lib$getdvi (&DviFullDevNam, 0, &DeviceNameDsc,
                   0, &StringItemDsc, &StringItemLength)))
      return (status);

   diptr->FullDevName[StringItemLength] = '\0';
   if (diptr->FullDevName[0] == '_')
   {
      /* copy the string over itself to get rid of the leading underscore */
      cptr = diptr->FullDevName;
      sptr = diptr->FullDevName+1;
      while (*sptr != '\0' && *sptr != ' ') *cptr++ = *sptr++;
      *cptr = '\0';
   }

   StringItemDsc.dsc$w_length = sizeof(diptr->MediaName);
   StringItemDsc.dsc$a_pointer = diptr->MediaName;

   if (VMSnok (status =
       lib$getdvi (&DviMediaName, 0, &DeviceNameDsc,
                   0, &StringItemDsc, &StringItemLength)))
      return (status);

   /* terminate the device name at the first white-space */
   diptr->MediaName[StringItemLength] = '\0';
   cptr = diptr->MediaName;
   while (*cptr != '\0' && *cptr != ' ') cptr++;
   *cptr = '\0';

   if (VMSnok (status =
       lib$getdvi (&DviVolSetMem, 0, &DeviceNameDsc,
                   &diptr->VolSetMem, 0, 0)))
      return (status);

   if (diptr->VolSetMem)
   {
      if (VMSok (status))
         status = lib$getdvi (&DviVolNumber, 0, &DeviceNameDsc,
                              &diptr->VolNumber, 0, 0);

      StringItemDsc.dsc$w_length = sizeof(diptr->NextDevName);
      StringItemDsc.dsc$a_pointer = diptr->NextDevName;

      status = lib$getdvi (&DviNextDevNam, 0, &DeviceNameDsc,
                           0, &StringItemDsc, &StringItemLength);

      diptr->NextDevName[StringItemLength] = '\0';
      if (diptr->NextDevName[0] == '_')
      {
         /* copy the string over itself to get rid of the leading underscore */
         cptr = diptr->NextDevName;
         sptr = diptr->NextDevName+1;
         while (*sptr != '\0' && *sptr != ' ') *cptr++ = *sptr++;
         *cptr = '\0';
      }
   }

   if (VMSnok (status =
       lib$getdvi (&DviMaxBlock, 0, &DeviceNameDsc,
                   &diptr->MaxBlock, 0, 0)))
      return (status);

   if (VMSnok (status =
       lib$getdvi (&DviFreeBlocks, 0, &DeviceNameDsc,
                   &diptr->FreeBlocks, 0, 0)))
      return (status);

   diptr->UsedBlocks = diptr->MaxBlock - diptr->FreeBlocks;

   if (Debug)
      fprintf (stdout,
"**********\n\
UcbStatus.lo32: %08.08X\n\
UcbStatus.hi32: %08.08X\n\
DevClass: %d\n\
DevChar: %d\n\
DevChar2: %d\n\
MaxBlock: %d\n\
FreeBlocks: %d\n\
UsedBlocks: %d\n\
VolNumber: %d\n\
VolSetMem: %d\n\
ErrCnt: %d\n\
FullDevName: %s\n\
NextDevName: %s\n\
MediaName: %s\n\
**********\n",
     diptr->UcbStatus.lo32,
     diptr->UcbStatus.hi32,
     diptr->DevClass,
     diptr->DevChar,
     diptr->DevChar2,
     diptr->MaxBlock,
     diptr->FreeBlocks,
     diptr->UsedBlocks,
     diptr->VolNumber,
     diptr->VolSetMem,
     diptr->ErrCnt,
     diptr->FullDevName,
     diptr->NextDevName,
     diptr->MediaName);

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
*/

ErrorGeneral
(
char *Text,
char *SourceFileName,
int SourceLineNumber
)
{
   register char  *cptr;

   /*********/
   /* begin */
   /*********/

   /* 
      The source file format provided by the "__FILE__" macro will be
      "device:[directory]name.type;ver".  Reduce that to "name.type".
   */
   for (cptr = SourceFileName; *cptr && *cptr != ';'; cptr++);
   if (*cptr)
   {
      while (*cptr != '.') cptr--;
      *cptr-- = '\0';
   }
   while (*cptr != ']') cptr--;
   cptr++;

   if (!HttpHasBeenOutput) fputs (Http404Header, HttpOut);
   fprintf (HttpOut,
"<!-- SoftwareID: %s Module: %s Line: %d -->\n\
<H1>ERROR!</H1>\n\
<P>Reported by server.\n\
<P>%s\n",
   SoftwareID, cptr, SourceLineNumber, Text);

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
*/

ErrorVmsStatus
(
int StatusValue,
char *Text,
char *HiddenText,
char *SourceFileName,
int SourceLineNumber
)
{
   static char  Message [256];
   static $DESCRIPTOR (MessageDsc, Message);

   register char  *cptr;
   int  status;
   short int  Length;

   /*********/
   /* begin */
   /*********/

   if (VMSok (status = sys$getmsg (StatusValue, &Length, &MessageDsc, 1, 0))) 
   {
      Message[Length] = '\0';
      Message[0] = toupper(Message[0]);
   }
   else
      exit (status);

   /* 
      The source file format provided by the "__FILE__" macro will be
      "device:[directory]name.type;ver".  Reduce that to "name.type".
   */
   for (cptr = SourceFileName; *cptr && *cptr != ';'; cptr++);
   if (*cptr)
   {
      while (*cptr != '.') cptr--;
      *cptr-- = '\0';
   }
   while (*cptr != ']') cptr--;
   cptr++;

   if (!HttpHasBeenOutput) fputs (Http404Header, HttpOut);
   fprintf (HttpOut,
"<!-- SoftwareID: %s Module: %s Line: %d -->\n\
<H1>ERROR!</H1>\n\
<P>Reported by server.\n\
<P>%s ... <TT>%s</TT>\n\
<!-- %%X%08.08X \"%s\" -->\n",
   SoftwareID, cptr, SourceLineNumber, Message, Text, StatusValue, HiddenText);

   return (SS$_NORMAL);
}

/****************************************************************************/
/*
Does a case-insensitive, character-by-character string compare and returns 
true if two strings are the same, or false if not.  If a maximum number of 
characters are specified only those will be compared, if the entire strings 
should be compared then specify the number of characters as 0.
*/ 

boolean strsame
(
register char *sptr1,
register char *sptr2,
register int  count
)
{
   while (*sptr1 && *sptr2)
   {
      if (toupper (*sptr1++) != toupper (*sptr2++)) return (false);
      if (count)
         if (!--count) return (true);
   }
   if (*sptr1 || *sptr2)
      return (false);
   else
      return (true);
}

/*****************************************************************************/
