/*****************************************************************************/
/*
                              CGIplus_main.c

CGIplus count.

Uses a sophisticated single function for accessing CGI variables. The function
differentiates between standard and CGIplus environments and changes behaviour
accordingly.  The function also serves to block waiting for subsequent CGIplus
requests.


08-JUN-97  MGD  v1.0.0, initial development
*/
/*****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <ctype.h>

int  IsCgiPlus = 0;
char  *CgiPlusEofPtr = NULL;

/* required prototypes */
char* CgiVar (char*);

/*****************************************************************************/
/*
*/

main (int argc, char* argv[])

{
   char  *cptr;

   /*********/
   /* begin */
   /*********/

   IsCgiPlus = ((CgiPlusEofPtr = getenv("CGIPLUSEOF")) != NULL);

   do {

      /* with CGIplus this call will block waiting for the next request */
      cptr = CgiVar ("WWW_REMOTE_HOST");

      original_main (argc, argv);

      if (IsCgiPlus)
      {
         /* the CGIplus EOF must be an independant I/O record */
         fflush (stdout);
         fprintf (stdout, CgiPlusEofPtr);
         fflush (stdout);
         /* release memory allocated when reading CGIplus variables */
         CgiVar (NULL);
      }

   } while (IsCgiPlus);

   exit (1);
}

/*****************************************************************************/
/*
This function provides a simple, uniform mechanism for a C Language script to
access CGI variable values transparently in both the standard CGI and the
CGIplus environments.

It is completely self-contained. Simply call the function with the CGI
variable name value to be accessed and if that CGI variable exists a pointer
to the value will be returned.  Errors and non-existant variables return NULL
pointers. The values pointed to should NOT be modified in any way (copy the
value if this is required). 

For the CGIplus environment supplying an empty variable name ("") returns each
CGIplus variable string (line from SYS$COMMAND) in turn, with no more being
indicated by a NULL pointer.

At the end of a CGIplus request call the function with a NULL parameter to
release allocated memory and reset ready for processing the next request's
variables.

The CGIplus functionality works by initially reading all lines from
SYS$COMMAND until the terminating blank line is encountered. These lines are
stored as a series of contiguous null-terminated strings, terminated by an
empty string, in allocated memory. These strings are searched for the required
variable name and a pointer to its value returned if found.
*/

char* CgiVar (char *VarName)

{
#  define CGIPLUS_LINE_SIZE 1024

   static int  VarPtrSize = 0;
   static char  *NextPtr = NULL,
                *VarPtr = NULL;

   register char  *cptr, *sptr;

   /*********/
   /* begin */
   /*********/

   if (VarName == NULL)
   {
      /* end of request, release memory and reinitialize */
      if (VarPtr != NULL)
      {
         free (VarPtr);
         NextPtr = VarPtr = NULL;
      }
      return (NULL);
   }

   if (CgiPlusEofPtr == NULL)
   {
      /* standard CGI environment, get variable from environment */
      if (!VarName[0]) return (NULL);
      return (getenv(VarName));
   }

   if (VarPtr == NULL)
   {
      /* read lines containing CGIplus variables from <stdin> */
      int  LineLength;
      char  Line [CGIPLUS_LINE_SIZE];
      FILE  *CgiPlusIn;
                                                 
      VarPtrSize = 0;
      if ((CgiPlusIn = fopen ("CGIPLUSIN", "r")) == NULL)
         return (NULL);
      while (fgets (Line, sizeof(Line), CgiPlusIn) != NULL)
      {
         /* first empty line signals the end of CGIplus variables */
         if (Line[0] == '\n') break;
         LineLength = strlen(Line);
         if ((VarPtr = realloc (VarPtr, VarPtrSize+LineLength+2)) == NULL)
            return (NULL);
         memcpy (VarPtr+VarPtrSize, Line, LineLength+1);
         VarPtrSize += LineLength;
         VarPtr[VarPtrSize-1] = '\0';
      }
      fclose (CgiPlusIn);
      if (VarPtr == NULL) return (NULL);
   }

   if (!VarName[0])
   {
      /* return each stored line one at a time */
      if (NextPtr == NULL) NextPtr = VarPtr;
      if (!*(sptr = cptr = NextPtr)) return (NextPtr = NULL);
      while (*cptr) cptr++;
      NextPtr = cptr + 1;
      return (sptr);
   }

   /* search CGIplus variable strings for required variable */
   sptr = VarPtr;
   while (*sptr)
   {
      /* simple comparison between supplied and in-string variable names */
      cptr = VarName;
      while (*cptr && *sptr && *sptr != '=')
      {
         if (toupper(*cptr) != toupper(*sptr)) break;
         cptr++;
         sptr++;
      }
      /* if found then return a pointer to the value */
      if (!*cptr && *sptr == '=') return (sptr+1);
      /* skip to next variable string (if any) */
      while (*sptr) sptr++;
      sptr++;
   }
   /* not found */
   return (NULL);
}

/*****************************************************************************/

